/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "plugin_statistics.h"

#include <iostream>
#include <numeric>
#include <sstream>

#include "item_pixbuf.h"




namespace quimeleon_plugin {





const char* PluginStatistics::_name = "Statistics";




PluginStatistics::PluginStatistics(quimeleon::Control* control)
	: quimeleon::Plugin(control)
{
	use_area("statistics");
	use_pixbuf("statisticspixbuf");
}




void
PluginStatistics::activate(const Glib::ustring& arg)
{
	quimeleon::ItemArea* item_area = get_area("statistics");

	if (item_area) {
		item_area->show();
	} else {
		std::cerr << "Cannot find area 'statistics'\n";
	}
}



void
PluginStatistics::deactivate()
{
	quimeleon::ItemArea* item_area = get_area("statistics");

	if (item_area) {
		item_area->hide();
	} else {
		std::cerr << "Cannot find area 'statistics'\n";
	}
}



std::vector<int>
vote(unsigned int correct, unsigned int n, int difficulty)
{
	std::vector<int> votes(n);
	int max = rand() % 500 + 1;	// max > 0
//	int max_wrong = 0;

	for (std::vector<int>::size_type i = 0; i < n; i++) {
		if (i == correct) {
			votes[i] = max;
		} else {
			votes[i] = rand() % max;
//			if (votes[i] > max_wrong) {
//				max_wrong = votes[i];
//			}
		}
	}
/*	votes[correct] += rand() % (max_wrong * (11 - (10 * difficulty / 255)))
		- max_wrong;
*/
	return votes;
}


void
PluginStatistics::draw(const Glib::ustring& id)
{
	quimeleon::ItemPixbuf* item_pixbuf = get_pixbuf(id);

	if (item_pixbuf) {
		static int question_n = -1;
		static unsigned int n;
		static std::vector<int> votes;
		static int total;

		if (question_n != int(control()->question_n())) {
			question_n = control()->question_n();
			n = control()->answers_cnt();
			votes = vote(control()->correct(), n,
				255 * (question_n + 1) / control()->questions_cnt());
			total = std::accumulate(votes.begin(), votes.end(), 0);
		}


		Glib::RefPtr<Gdk::Drawable>& drawable = item_pixbuf->drawable();

		item_pixbuf->get_gc()->set_foreground(item_pixbuf->color_active());
		item_pixbuf->get_gc()->set_background(item_pixbuf->color_inactive());
		const int bar_height = item_pixbuf->height() / (n + 1);

		Glib::RefPtr<Pango::Layout> layout
			= Pango::Layout::create(item_pixbuf->pango_context());
		Pango::FontDescription font_description
			= Pango::FontDescription("Sans 14");
		font_description.set_size(
			int(item_pixbuf->font_scale() * font_description.get_size()));
//		layout->set_width(item_pixbuf->width() * 1000);
		layout->set_font_description(font_description);
		layout->set_alignment(Pango::ALIGN_CENTER);
		int layout_question_w;
		int layout_percent_w;
		int layout_h;
		layout->set_text("100%");
		layout->get_pixel_size(layout_percent_w, layout_h);
		layout->set_text("W");
		layout->get_pixel_size(layout_question_w, layout_h);


		for (std::vector<int>::size_type i = 0; i< votes.size(); i++) {
			const int bar_width
				=  (item_pixbuf->width() - layout_question_w - layout_percent_w)
				 * votes[i] / total;
			const int y = (i * item_pixbuf->height() + bar_height) / n
				- (bar_height / n / 2);
			item_pixbuf->get_gc()->set_foreground(item_pixbuf->color_active());
			drawable->draw_rectangle(item_pixbuf->get_gc(), true,
				layout_question_w, y,
				bar_width, bar_height);


			item_pixbuf->get_gc()->set_foreground(item_pixbuf->color_normal());
			char label[2] = { 'A' + i, '\0' };
			layout->set_text(label);
			drawable->draw_layout(item_pixbuf->get_gc(),
				0, y + bar_height / 2 - layout_h / 2, layout);

			std::ostringstream ostr;
			ostr << (100 * votes[i] / total) << '%';
			layout->set_text(ostr.str());
			drawable->draw_layout(item_pixbuf->get_gc(),
				bar_width + layout_question_w,
				y + bar_height / 2 - layout_h / 2,
				layout);
		}

		item_pixbuf->put_drawable();
	} else {
		std::cerr << "Area 'statistics' has no pixbuf\n";
	}
}





} // end namespace quimeleon_plugin
