/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2003, 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "audio.h"

#include <iostream>



#ifdef HAVE_SOUND

# include <glibmm/timer.h>


# include <cstdio>
# include <cstdlib>
# include <unistd.h>

# include <vorbis/codec.h>

#include "gettext.h"

#endif




namespace quimeleon {




#ifdef HAVE_SOUND


const int gl_buf_size = 4096;
char gl_audio_buf[gl_buf_size]; // use memory from data segment




Glib::Thread*
init_threads()
{
	if (! Glib::thread_supported()) {
		Glib::thread_init();
	}

	return 0;
}






int Audio::_driver = -1;
ao_device *Audio::_device = 0;

int Audio::_channels = 0;
int Audio::_rate = 0;

OggVorbis_File Audio::_oggvorbis_file;

Glib::Thread* Audio::_thread = init_threads();

bool Audio::_is_thread_running = false;
bool Audio::_is_oggvorbis = false;

int Audio::_initialized = 0;

Glib::Mutex Audio::_mutex_common_data;


#endif





Audio::Audio()
{
#ifdef HAVE_SOUND
	if (! Glib::thread_supported()) {
		Glib::thread_init();
	}

	Glib::Mutex::Lock lock(_mutex_common_data);
	if (! _initialized) {
		ao_initialize();
	}
	_initialized++;

#endif
}



Audio::~Audio()
{
#ifdef HAVE_SOUND
	_mutex_common_data.lock();

	if (_is_thread_running) {
		if (_is_oggvorbis) {

			ov_clear(&_oggvorbis_file);
			_is_oggvorbis = false;
		}
		_mutex_common_data.unlock();

		_thread->join();
	} else {
		_mutex_common_data.unlock();
	}


	_initialized--;
	if (! _initialized) {
		ao_shutdown();
	}
#endif
}




void
Audio::reset()
{
#ifdef HAVE_SOUND
	_audio_files.clear();
#endif
}



bool
Audio::play(audio_t audio)
{
#ifdef HAVE_SOUND
	const Glib::ustring filename = _audio_files[audio];
	if (! filename.empty()) {
		return play(filename);
	}
#endif

	return false;
}



bool
Audio::play(const Glib::ustring& filename)
{
#ifdef HAVE_SOUND
	try {
		Glib::Mutex::Lock lock(_mutex_common_data);

		if (   load_file(filename)
		    && ! _is_thread_running) {
			_thread = Glib::Thread::create(sigc::ptr_fun(&Audio::play_thread),
			                               true);
			_thread->set_priority(Glib::THREAD_PRIORITY_LOW);
		}

	} catch (const Glib::Error& e) {
		std::cerr << e.what() << '\n';
	}
#endif

	return false;
}



void
Audio::mute()
{
#ifdef HAVE_SOUND
	Glib::Mutex::Lock lock(_mutex_common_data);

	if (_is_oggvorbis) {
		ov_clear(&_oggvorbis_file);
		_is_oggvorbis = false;
	}
#endif
}



void
Audio::set_sound(audio_t audio, Glib::ustring file)
{
#ifdef HAVE_SOUND
	_audio_files[audio] = file;
#endif
}



#ifdef HAVE_SOUND
// _mutex_common_data to be locked by caller
bool
Audio::load_file(const Glib::ustring& filename) throw()
{
	FILE* fh = fopen(filename.c_str(), "r");

	if (fh == 0) {
		return false;
	}


	if (_is_oggvorbis) {
		ov_clear(&_oggvorbis_file);
		_is_oggvorbis = false;
	}

	if (ov_open(fh, &_oggvorbis_file, 0, 0) < 0) {
		return false;
	}
	_is_oggvorbis = true;


	vorbis_info* vi = ov_info(&_oggvorbis_file, -1);

	bool is_new_device = false;
	if (   ! _device
	    || vi->channels != _channels
	    || vi->rate != _rate) {
		is_new_device = true;
		if (_device) {
			ao_close(_device);
			_device = 0;
		}

		ao_sample_format format;
		format.bits = 16;
		format.byte_format = AO_FMT_LITTLE;
		format.channels = vi->channels;
		format.rate = vi->rate;

		// first try the last working driver, then the default driver, if
		// that fails try drivers that can handle output by multiple clients
		if (   _driver == -1
		    || ! (_device = ao_open_live(_driver, &format, 0))) {
			_driver = ao_default_driver_id();
			if (   _driver == -1
			    || ! (_device = ao_open_live(_driver, &format, 0))) {
				_driver = ao_driver_id("alsa09");
				if (   _driver == -1
				    || ! (_device = ao_open_live(_driver, &format, 0))) {
					_driver = ao_driver_id("arts");
					if (   _driver == -1
					    || ! (_device = ao_open_live(_driver, &format, 0))) {
						_driver = ao_driver_id("esd");
						if (   _driver == -1
						   || ! (_device = ao_open_live(_driver, &format, 0))) {
							std::cerr << _("No suitable audio driver found")
								<< '\n';
						}
					}
				}
			}
		}

		if (_device) {
			_channels = format.channels;
			_rate = format.rate;
		}
	}


	if (! _device) {
		ov_clear(&_oggvorbis_file);

		return false;
	}

	return is_new_device;
}



void
Audio::play_thread()
{
	{
		Glib::Mutex::Lock lock(_mutex_common_data);
		_is_thread_running = true;
	}

	int bitstream;	// out parameter of ov_read
	long len;
	for (;;) {
		{
			Glib::Mutex::Lock lock(_mutex_common_data);

			if (! _is_oggvorbis) {
				break;
			}
			len = ov_read(&_oggvorbis_file, gl_audio_buf, gl_buf_size,
			              0, 2, 1,	// little endian, 16-bit, signed
			              &bitstream);

			if (len <= 0) {	// eof or error
				break;
			} else if (! ao_play(_device, gl_audio_buf, len)) {
				break;
			}
		}
		Glib::usleep(1000);
	}

	{
		Glib::Mutex::Lock lock(_mutex_common_data);
		if (_is_oggvorbis) {
			ov_clear(&_oggvorbis_file);
			_is_oggvorbis = false;
		}
		if (_device) {
			ao_close(_device);
			_device = 0;
		}

		_is_thread_running = false;
	}
}
#endif





} // end namespace quimeleon
