/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2003, 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "configuration.h"

#include <fstream>
#include <limits>

#include <climits>
#include <cstdlib>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


#include <gtkmm/messagedialog.h>

#include "item_action.h"
#include "plugin.h"

#include "gettext.h"




namespace quimeleon {





const char* c_default_datadir_conf = "datadir.conf";
const char* c_default_options_conf = "options.conf";




std::string
absolute_path(const std::string& filename)
{
	if (! Glib::path_is_absolute(filename)) {
		return Glib::build_filename(Glib::get_current_dir(), filename);
	}


	return filename;
}



void
dir_must_exist(const char* pathname)
{
	struct stat info;
	if (stat(pathname, &info) == -1) {
		if (mkdir(pathname, 0750) == -1) {
			throw CreateError(pathname);
		}
	} else if (! S_ISDIR(info.st_mode)) {
		throw CreateError(pathname);
	}
}




Configuration::Configuration()
	: _trivia_file("default.trivia"),
	  _style_file("default.style"),
	  _max_level_cnt(std::numeric_limits<unsigned int>::max()),
	  _do_buffer(true),
	  _fullscreen(false),
	  _do_show_hint(false)
{
	std::string locale_name = std::locale().name();
	if (locale_name.size() > 2) {
		_lang = locale_name.substr(0, 2);
	}
	_item_top = new ItemTop;
	_audio = new Audio;

	add_data_dir(DATADIRROOT);
	add_data_dir(Glib::build_filename(DATADIRROOT, "quimeleon"));
	add_data_dir(Glib::build_filename(DATADIRROOT, "default"));


	std::vector<std::string> path(3);
	path[0] = SYSCONFDIR;
	path[1] = PACKAGE;
	path[2] = c_default_datadir_conf;
	std::string global_config_file = Glib::build_filename(path);
	read_datadir_file(global_config_file);

# ifndef NODOTPREFIX
	path[1] = "." PACKAGE;
# endif

#ifdef GROUPDIRENV
	bool is_found;
	std::string env_group = Glib::getenv(GROUPDIRENV, is_found);
	if (is_found) {
		path[0] = env_group;
		std::string group_config_file = Glib::build_filename(path);
		read_datadir_file(group_config_file);
	}
#endif

	_global_data_dirs_size = _data_dirs.size();
	_global_trivia_file = _trivia_file;
	_global_style_file = _style_file;
	_global_style_name = _style_name;
	_global_do_buffer = _do_buffer;
	_global_fullscreen = _fullscreen;
	_global_do_show_hint = _do_show_hint;


	std::string env_home = Glib::get_home_dir();
	if (! env_home.empty()) {
		path[0] = env_home;
		std::string user_config_file = Glib::build_filename(path);
		read_datadir_file(user_config_file);
	}

}



Configuration::Configuration(const Configuration& configuration)
{
	*this = configuration;
	_item_top = new ItemTop;
	_audio = new Audio;
}



Configuration::~Configuration()
{
	delete _item_top;
	delete _audio;
}




void
Configuration::operator=(const Configuration& configuration)
{
	_lang = configuration._lang;
	_data_dirs = configuration._data_dirs;
	_trivia_file = configuration._trivia_file;
	_style_file = configuration._style_file;
	_style_name = configuration._style_name;
	_max_level_cnt = configuration._max_level_cnt;
	_do_buffer = configuration._do_buffer;
	_fullscreen = configuration._fullscreen;
	_do_show_hint = configuration._do_show_hint;
	_global_data_dirs_size = configuration._global_data_dirs_size;
	_global_trivia_file = configuration._global_trivia_file;
	_global_style_file = configuration._global_style_file;
	_global_style_name = configuration._global_style_name;
	_global_do_buffer = configuration._global_do_buffer;
	_global_fullscreen = configuration._global_fullscreen;
}




void
Configuration::add_data_dir(std::string dirname)
{
	if (dirname.empty()) {
		return;
	}

	struct stat info;
	if (   stat(dirname.c_str(), &info) == -1
	    || ! S_ISDIR(info.st_mode)) {
		return;
	}


	if (G_IS_DIR_SEPARATOR(dirname[dirname.length() - 1])) {
		_data_dirs.push_back(dirname.substr(0, dirname.length() - 1));
	} else {
		_data_dirs.push_back(dirname);
	}
}



void
Configuration::reset_style()
{
	_item_top->reset();
	_audio->reset();
}



std::vector<std::string>
Configuration::plugins()
{
	std::vector<std::string> re;

	const std::vector<ItemAction*>& actions = _item_top->all_actions();

	for (std::vector<ItemAction*>::const_iterator i = actions.begin();
	     i < actions.end(); i++) {
		const std::string& filename = (*i)->plugin();
		if (! filename.empty()) {
			re.push_back(filename);
		}
	}

	std::copy(_action_autonomouss.begin(), _action_autonomouss.end(),
		std::back_inserter(re));

	return re;
}



std::string
Configuration::find_file(const std::string& filename) const
{
	struct stat info;

	if (stat(filename.c_str(), &info) != -1) {
		return filename;
	}


	if (! _style_name.empty()) {
		std::vector<std::string> path(4);
		path[0] = DATADIRROOT;
		path[1] = _style_name;
		path[2] = _lang;
		path[3] = filename;
		std::string re = Glib::build_filename(path);
		if (stat(re.c_str(), &info) != -1) {
			return re;
		}
		path.resize(3);
		path[2] = filename;
		re = Glib::build_filename(path);
		if (stat(re.c_str(), &info) != -1) {
			return re;
		}
	}

	std::vector<std::string> path3(3);
	std::vector<std::string> path2(2);
	path3[1] = _lang;
	path3[2] = path2[1] = filename;
	for (std::vector<std::string>::const_reverse_iterator i
	     	= _data_dirs.rbegin();
	     i < _data_dirs.rend(); i++) {
		std::vector<std::string> path(3);
		path3[0] = *i;
		std::string re = Glib::build_filename(path3);
		if (stat(re.c_str(), &info) != -1) {
			return re;
		}
		path2[0] = *i;
		re = Glib::build_filename(path2);
		if (stat(re.c_str(), &info) != -1) {
			return re;
		}
	}


	path2[0] = PLUGINDIR;
	const std::string re = Glib::build_filename(path2);
	if (stat(re.c_str(), &info) != -1) {
		return re;
	}


	throw FileNotFound(filename);
}




void
Configuration::read_datadir_file(const std::string& filename)
{
	std::ifstream in(filename.c_str());
	if (in) {
		std::string path;
		while (! in.eof()) {
			std::getline(in, path);
			if (   path.empty()
			    || path[0] == '#') {
				continue;
			}
			add_data_dir(path);
		}
	}
}



void
verify_key(const Glib::ustring& expected, const Glib::ustring& key)
{
	if (expected != key) {
		throw ParserError(_("unknown key") + ' ' + key);
	}
}




bool
Configuration::read_options_file(const std::string& filename)
{
	std::ifstream in(filename.c_str());
	if (in) {
		std::string key_value;
		while (! in.eof()) {
			std::getline(in, key_value);
			if (key_value.empty()) {
				continue;
			}
			std::string::size_type pos = key_value.find('=');
			Glib::ustring key = key_value.substr(0, pos);
	
			Glib::ustring value;
			if (pos != std::string::npos) {
				value = key_value.substr(pos + 1);
			}
	
			switch (key[0]) {
			case 's':	// style + stylefile
				if (key.length() > 5) {
					verify_key("stylefile", key);
					set_style_file(value);
				} else {
					verify_key("style", key);
					set_style_name(value);
				}
				break;
			case 't':	// triviafile
				verify_key("triviafile", key);
				set_trivia_file(value);
				break;
			case 'n':	// no-buffer
				verify_key("no-buffer", key);
				set_do_buffer(false);
				break;
			case 'b':	// buffer
				verify_key("buffer", key);
				set_do_buffer(true);
				break;
			case 'f':	// fullscreen
				verify_key("fullscreen", key);
				if (   value.empty()
				    || value == "true") {
					set_fullscreen(true);
				} else {
					set_fullscreen(false);
				}
				break;
			case 'h':	// hint
				verify_key("hint", key);
				if (   value.empty()
				    || value == "true") {
					set_do_show_hint(true);
				} else {
					set_do_show_hint(false);
				}
				break;
			case '#':
			case '[':
				break;
			default:
				throw ParserError(_("unknown key") + ' ' + key);
			}
		}

		return true;
	} else {	// if (in)
		return false;
	}
}



void
Configuration::read_options()
{
	std::vector<std::string> path(3);
	path[1] =
#ifndef NODOTPREFIX
	           "." PACKAGE;
#else
	           PACKAGE;
#endif
	path[2] = c_default_options_conf;


#ifdef GROUPDIRENV
	bool is_found;
	path[0] = Glib::getenv(GROUPDIRENV, is_found);
	if (is_found) {
		std::string group_options_file = Glib::build_filename(path);
		read_options_file(group_options_file);
	}
#endif



	path[0] = Glib::get_home_dir();
	if (! path[0].empty()) {
		std::string user_options_file = Glib::build_filename(path);
		if (! read_options_file(user_options_file)) {
			set_do_show_hint(true);
		}
	}
}




enum option_index_t {
	i_triviafile,
	i_stylefile,
	i_style,
	i_fullscreen,
	i_buffer,
	i_show_hint,
	i_size
};

void
Configuration::save_style_and_trivia()
{
	std::vector<std::string> path(2);

	path[0] = Glib::get_home_dir();
	if (! path[0].empty()) {
		path[1] =
#ifndef NODOTPREFIX
		          "." PACKAGE;
#else
		          PACKAGE;
# endif

		std::string user_config_dir = Glib::build_filename(path);
		dir_must_exist(user_config_dir.c_str());


		std::map<std::string, std::vector<std::string> > path_comments;
		std::vector<std::string> intro_comment;
		std::vector<std::string> lines;

		for (std::vector<std::string>::size_type i = _global_data_dirs_size;
		     i < _data_dirs.size(); i++) {
			path_comments.insert(
				std::pair<std::string, std::vector<std::string> >(
					_data_dirs[i], std::vector<std::string>()));
		}

		path.push_back(c_default_datadir_conf);
		std::string user_config_file = Glib::build_filename(path);
		std::ifstream in(user_config_file.c_str());
		if (in) {
			std::string line;

			while (! in.eof()) {
				std::getline(in, line);

				if (line.empty()) {
					intro_comment = lines;
					lines.clear();
					break;
				}
				if (line[0] == '#') {
					lines.push_back(line);
				} else {
					if (G_IS_DIR_SEPARATOR(line[line.length() - 1])) {
						line.resize(line.length() - 1);
					}
					std::map<std::string, std::vector<std::string> >::iterator i
						= path_comments.find(line);
					if (i != path_comments.end()) {
						i->second = lines;
					} else {
						intro_comment = lines;
					}
					lines.clear();
					break;
				}
			}

			while (! in.eof()) {
				std::getline(in, line);

				if (   line.empty()
				    || line[0] == '#') {
					lines.push_back(line);
				} else {
					if (G_IS_DIR_SEPARATOR(line[line.length() - 1])) {
						line.resize(line.length() - 1);
					}
					std::map<std::string, std::vector<std::string> >::iterator i
						= path_comments.find(line);
					if (i != path_comments.end()) {
						i->second = lines;
					}
					lines.clear();
				}
			}

			in.close();
		}


		std::ofstream out(user_config_file.c_str());
		if (out) {
			if (! intro_comment.empty()) {
				for (std::vector<std::string>::iterator i
				     	= intro_comment.begin();
				     i < intro_comment.end(); i++) {
					out << *i << '\n';
				}
				out << '\n';
			}

			for (std::vector<std::string>::size_type i = _global_data_dirs_size;
			     i < _data_dirs.size(); i++) {
				std::map<std::string, std::vector<std::string> >::iterator j
					= path_comments.find(_data_dirs[i]);
				if (j != path_comments.end()) {
					for (std::vector<std::string>::iterator k
					     	= j->second.begin();
					     k < j->second.end(); k++) {
						out << *k << '\n';
					}
				}
				out << _data_dirs[i] << '\n';
			}

			if (   ! lines.empty()
			    && lines.back().empty()) {
				lines.pop_back();
			}
			for (std::vector<std::string>::iterator i = lines.begin();
			     i < lines.end(); i++) {
				out << *i << '\n';
			}
		} else {
			throw WriteError(user_config_file);
		}


		std::vector<bool> done_options(i_size, false);
		lines.clear();

		path[2] = c_default_options_conf;
		std::string user_options_file = Glib::build_filename(path);

		std::ifstream options_in(user_options_file.c_str());
		if (options_in) {
			std::string line;

			while (! options_in.eof()) {
				std::getline(options_in, line);

				if (   line.empty()
				    || line[0] == '[') {
					lines.push_back(line);
				} else {
					std::string::size_type pos = line.find('=');

					if (pos == std::string::npos) {
						lines.push_back(line);
						continue;
					}

					std::string key;
					if (line[0] == '#') {
						key = line.substr(1, pos - 1);
					} else {
						key = line.substr(0, pos);
					}

					std::string value = line.substr(pos + 1);
					bool do_comment_out = false;

					try {
					switch (key[0]) {
					case 't':	// triviafile
						verify_key("triviafile", key);
						if (_trivia_file != _global_trivia_file) {
							value = _trivia_file;
						} else if (_trivia_file != value) {
							do_comment_out = true;
						}
						done_options[i_triviafile] = true;
						break;
					case 's':	// style + stylefile
						if (key.length() > 5) {
							verify_key("stylefile", key);
							if (_style_file != _global_style_file) {
								value = _style_file;
							} else if (_style_file != value) {
								do_comment_out = true;
							}
							done_options[i_stylefile] = true;

							break;
						}
					default:
						lines.push_back(line);
						continue;
					}
					} catch (const ParserError& e) {
						lines.push_back(line);
						continue;
					}

					if (do_comment_out) {
						if (line[0] == '#') {
							lines.push_back(line);
						} else {
							lines.push_back(std::string("#") + line);
						}
					} else {
						lines.push_back(key + '=' + value);
					}
				}
			}

			options_in.close();
		} // end if (in)



		std::ofstream options_out(user_options_file.c_str());
		if (options_out) {
			if (   ! lines.empty()
			    && lines.back().empty()) {
				lines.pop_back();
			}

			for (std::vector<std::string>::const_iterator i = lines.begin();
			     i < lines.end(); i++) {
				options_out << *i << '\n';
			}
			for (int i = i_triviafile; i <= i_stylefile; i++) {
				switch (i) {
				case i_triviafile:
					if (   ! done_options[i_triviafile]
					    && _trivia_file != _global_trivia_file) {
						options_out << "triviafile=" << _trivia_file << '\n';
					}
					break;
				case i_stylefile:
					if (   ! done_options[i_stylefile]
					    && _style_file != _global_style_file) {
						options_out << "stylefile=" << _style_file << '\n';
					}
					break;
				}
			}
		} else {
			throw WriteError(user_options_file);
		}
	} // end if (! path[0].empty())

}



void
Configuration::save_options()
{
	std::vector<std::string> path(2);

	path[0] = Glib::get_home_dir();
	if (! path[0].empty()) {
		std::vector<std::string> lines;
		std::vector<bool> done_options(i_size, false);

		path[1] =
#ifndef NODOTPREFIX
		          "." PACKAGE;
#else
		          PACKAGE;
# endif

		std::string user_config_dir = Glib::build_filename(path);
		dir_must_exist(user_config_dir.c_str());

		path.push_back(c_default_options_conf);
		std::string user_options_file = Glib::build_filename(path);

		std::ifstream in(user_options_file.c_str());
		if (in) {
			std::string line;

			while (! in.eof()) {
				std::getline(in, line);

				if (   line.empty()
				    || line[0] == '[') {
					lines.push_back(line);
				} else {
					std::string::size_type pos = line.find('=');

					std::string key;
					std::string value;

					if (line[0] == '#') {
						key = line.substr(1, pos - 1);
					} else {
						key = line.substr(0, pos);
					}
					if (pos != std::string::npos) {
						value = line.substr(pos + 1);
					}
					bool do_comment_out = false;

					try {
					switch (key[0]) {
					case 's':	// style + stylefile
						if (key.length() > 5) {
							verify_key("stylefile", key);
							if (_style_file != _global_style_file) {
								value = _style_file;
							} else if (_style_file != value) {
								do_comment_out = true;
							}
							done_options[i_stylefile] = true;
						} else {
							verify_key("style", key);
							if (_style_name != _global_style_name) {
								value = _style_name;
							} else if (_style_name != value) {
								do_comment_out = true;
							}
							done_options[i_style] = true;
						}
						break;
					case 't':	// triviafile
						verify_key("triviafile", key);
						if (_trivia_file != _global_trivia_file) {
							value = _trivia_file;
						} else if (_trivia_file != value) {
							do_comment_out = true;
						}
						done_options[i_triviafile] = true;
						break;
					case 'n':	// no-buffer
						verify_key("no-buffer", key);
						if (_do_buffer != _global_do_buffer) {
							if (_do_buffer) { // i.e. _global_do_buffer == false
								key = "buffer";	// substitute with "buffer"
							}
						} else if (_do_buffer) {
							do_comment_out = true;
						}
						done_options[i_buffer] = true;
						break;
					case 'b':	// buffer
						verify_key("buffer", key);
						if (_do_buffer != _global_do_buffer) {
							if (! _do_buffer) {	// i.e. _global_do_buffer == true
								// substitute with "no-buffer"
								key = "no-buffer";
							}
						} else if (! _do_buffer) {
							do_comment_out = true;
						}
						done_options[i_buffer] = true;
						break;
					case 'f':	// fullscreen
						verify_key("fullscreen", key);
						if (_fullscreen != _global_fullscreen) {
							if (   value.empty()
							    || value == "true") { // i.e. _global_fullscreen = false
								if (! _fullscreen) {
									// substitute with "fullscreen=false"
									value = "false";
								}
							} else if (_fullscreen) {
								// substitute with "fullscreen"
								value.clear();
							}
						} else if (_fullscreen) {
							if (   ! value.empty()
							    && value != "true") {
								do_comment_out = true;
							}
						} else if (   value.empty()
						           || value == "true") {
							do_comment_out = true;
						}
						done_options[i_fullscreen] = true;
						break;
					case 'h':	// hint
						verify_key("hint", key);
						if (_do_show_hint != _global_do_show_hint) {
							if (   value.empty()
							    || value == "true") { // i.e. _global_do_show_hint = false
								if (! _do_show_hint) {
									// substitute with "do_show_hint=false"
									value = "false";
								}
							} else if (_do_show_hint) {
								// substitute with "do_show_hint"
								value.clear();
							}
						} else if (_do_show_hint) {
							if (   ! value.empty()
							    && value != "true") {
								do_comment_out = true;
							}
						} else if (   value.empty()
						           || value == "true") {
							do_comment_out = true;
						}
						done_options[i_show_hint] = true;
						break;
					default:
						do_comment_out = true;
					} // end switch (key[0])
					} catch (const ParserError& e) {
						if (line[0] == '#') {
							lines.push_back(line);
							continue;
						} else {
							throw;
						}
					}

					if (do_comment_out) {
						if (line[0] == '#') {
							lines.push_back(line);
						} else {
							lines.push_back(std::string("#") + line);
						}
					} else if (value.empty()) {
						lines.push_back(key);
					} else {
						lines.push_back(key + '=' + value);
					}
				}
			}

			in.close();
		} // end if (in)



		std::ofstream out(user_options_file.c_str());
		if (out) {
			if (   ! lines.empty()
			    && lines.back().empty()) {
				lines.pop_back();
			}

			for (std::vector<std::string>::const_iterator i = lines.begin();
			     i < lines.end(); i++) {
				out << *i << '\n';
			}


			for (int i = 0; i < i_size; i++) {
				switch (i) {
				case i_triviafile:
					if (   ! done_options[i_triviafile]
					    && _trivia_file != _global_trivia_file) {
						out << "triviafile=" << _trivia_file << '\n';
					}
					break;
				case i_stylefile:
					if (   ! done_options[i_stylefile]
					    && _style_file != _global_style_file) {
						out << "stylefile=" << _style_file << '\n';
					}
					break;
				case i_style:
					if (   ! done_options[i_style]
					    && _style_name != _global_style_name) {
						out << "style=" << _style_name << '\n';
					}
					break;
				case i_fullscreen:
					if (   ! done_options[i_fullscreen]
					    && _fullscreen != _global_fullscreen) {
						if (_fullscreen) {
							out << "fullscreen\n";
						} else {
							out << "fullscreen=false\n";
						}
					}
					break;
				case i_buffer:
					if (   ! done_options[i_buffer]
					    && _do_buffer != _global_do_buffer) {
						if (_do_buffer) {
							out << "buffer\n";
						} else {
							out << "no-buffer\n";
						}
					}
					break;
				case i_show_hint:
					if (   ! done_options[i_show_hint]
					    && _do_show_hint != _global_do_show_hint) {
						if (_do_show_hint) {
							out << "hint\n";
						} else {
							out << "hint=false\n";
						}
					}
					break;
				}
			}
		} else {
			throw WriteError(user_options_file);
		}
	}
}





} // end namespace quimeleon
