/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "control.h"

#include "item_action.h"
#include "item_level.h"
#include "item_answer.h"
#include "item_pixmap.h"
#include "item_status.h"
#include "plugin.h"

#include "gettext.h"




namespace quimeleon {





Control::Control(Configuration* configuration,
	const Trivias& trivias)
	: _configuration(configuration),
	  _trivias(trivias),
	  _level_cnt(trivias.max_level_cnt()),
	  _state(STATE_black),
	  _substate(SUBSTATE_normal),
	  _question_n(0),
	  _level_n(0),
	  _action_active(0),
	  _area_output_overlay(_configuration->item_top()->get_area("output_overlay")),
	  _area_output(_configuration->item_top()->get_area("output"))
{
	_level_cnt = _configuration->item_top()->set_level_cnt(_level_cnt);

	if (_area_output_overlay) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output_overlay->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_do_buffer(false);
		}
	}
	if (_area_output) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_do_buffer(false);
		}
	}
}




void
Control::on_item_top_changed()
{
	ItemTop* top = _configuration->item_top();
	_level_cnt = top->set_level_cnt(_level_cnt);

	_area_output_overlay = top->get_area("output_overlay");
	if (_area_output_overlay) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output_overlay->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_do_buffer(false);
		}
	}
	_area_output = top->get_area("output");
	if (_area_output) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_do_buffer(false);
		}
	}

	for (unsigned int i = 0; i < _level_n; i++) {
		top->mark_state_level(i, active);
	}
	top->mark_state_level(_level_n, selected);

	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
		 i < statuss.end(); i++) {
		(*i)->set_level(_level_n);
	}

	if (   _state >= STATE_question
	    && _state <= STATE_show_answer) {
		set_question(question_n());
	}
}



void
Control::set_trivia(const Trivias& trivias)
{
	_trivias = trivias;
	_level_cnt = _configuration->item_top()->set_level_cnt(
		_trivias.max_level_cnt());

	reset_level();

	if (   _state >= STATE_question
	    && _state <= STATE_show_answer) {
		set_question(0);
	}
}




void
Control::set_gamestate(gamestate_t gamestate) throw()
{
	_state = gamestate;
}



gamestate_t
Control::gamestate() throw()
{
	return _state;
}



ItemArea*
Control::get_area(const Glib::ustring& id)
{
	return _configuration->item_top()->get_area(id);
}



unsigned int
Control::questions_cnt() const
{
	return _trivias.size();
}



unsigned int
Control::question_n() const throw()
{
	return _question_n;
}



bool
Control::set_question(unsigned int n)
{
	if (   ! _trivias.empty()
	    && n < _trivias.size()) {
		_question_n = n;
		free_canvas();
		if (   _area_output
		    && _area_output->visible()) {
			_area_output->hide();
		} else if (   _area_output_overlay
		           && _area_output_overlay->visible()) {
			_area_output_overlay->hide();
		}

		_state = STATE_question;
		_substate = SUBSTATE_normal;

		bool alternative_changed;
		_correct_pos = _configuration->item_top()->set_question(
				_trivias[_question_n].question(),
				_trivias[_question_n].answers(),
				_trivias[_question_n].correct(),
				alternative_changed);

		const std::vector<ItemStatus*>& statuss =
			_configuration->item_top()->statuss();
		for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
		     i < statuss.end(); i++) {
			(*i)->set_question(_question_n);
		}

		if (alternative_changed) {
			_area_output_overlay =
				_configuration->item_top()->get_area("output_overlay");
			if (_area_output_overlay) {
				ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
					_area_output_overlay->find_id("image"));

				if (item_pixmap) {
					item_pixmap->set_do_buffer(false);
				}
			}
			_area_output = _configuration->item_top()->get_area("output");
			if (_area_output) {
				ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
					_area_output->find_id("image"));

				if (item_pixmap) {
					item_pixmap->set_do_buffer(false);
				}
			}
		}

		if (_trivias[_question_n].question_audio().empty()) {
			_configuration->audio()->play(Audio::question_new);
		} else {
			Audio::play(_trivias[_question_n].question_audio());
		}

		if (! _trivias[_question_n].question_image().empty()) {
			image_overlay(_trivias[_question_n].question_image());
		}

		return true;
	}

	return false;
}



void
Control::image_overlay_prepare(const Glib::ustring& image)
{
	_image_overlaid = image;

	if (SUBSTATE_overlay && _area_output_overlay) {
		_area_output_overlay->set_visible(false);
	}

	if (_area_output) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_normal(_image_overlaid);
			_area_output->show();
		} else {
			std::cerr << "<item type=\"pixmap\" id=\"image\" /> "
				<< _("not found") << '\n';
		}
	}

	if (_area_output_overlay) {
		_substate = SUBSTATE_overlay_prepared;
	}
}



void
Control::image_overlay(const Glib::ustring& image)
{
	_image_overlaid = image;

	if (SUBSTATE_overlay_prepared && _area_output) {
		_area_output->hide();
	}


	if (_area_output_overlay) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output_overlay->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_normal(_image_overlaid);
			_area_output_overlay->show();

			_substate = SUBSTATE_overlay;
		} else {
			std::cerr << "<item type=\"pixmap\" id=\"image\" /> "
				<< _("not found") << '\n';
		}
	} else if (_area_output) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_normal(_image_overlaid);
			_area_output->show();
		} else {
			std::cerr << "<item type=\"pixmap\" id=\"image\" /> "
				<< _("not found") << '\n';
		}

		_substate = SUBSTATE_normal;
	}

}



void
Control::image_shrink()
{
	if (_substate != SUBSTATE_overlay) {
		return;
	}

	_area_output_overlay->hide();
	if (_area_output) {
		ItemPixmap* item_pixmap = dynamic_cast<ItemPixmap*>(
			_area_output->find_id("image"));

		if (item_pixmap) {
			item_pixmap->set_normal(_image_overlaid);
			_area_output->show();
		} else {
			std::cerr << "<item type=\"pixmap\" id=\"image\" /> "
				<< _("not found") << '\n';
		}
	}

	_substate = SUBSTATE_normal;
}



bool
Control::substate_next()
{
	switch (_substate) {
	case SUBSTATE_overlay_prepared:
		image_overlay(_image_overlaid);

		return true;
	case SUBSTATE_overlay:
		image_shrink();

		return true;
	default:
		return false;
	}
}



bool
Control::prev_question()
{
	if (   _state >= STATE_question
	    && _state <= STATE_show_answer
	    && _question_n > 0) {
		return set_question(_question_n - 1);
	}

	return false;
}



bool
Control::next_question()
{
	if (   _state >= STATE_question
	    && _state <= STATE_show_answer
	    && _question_n + 1 < _trivias.size()) {
		const std::vector<ItemStatus*>& statuss =
			_configuration->item_top()->statuss();
		for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
		     i < statuss.end(); i++) {
			(*i)->next_question();
		}
		return set_question(_question_n + 1);
	}

	return false;
}



void
Control::reset_question()
{
	static const std::vector<answer_t> answers_empty;
	bool alternative_changed;
	_configuration->item_top()->set_question(
		"", answers_empty, 0, alternative_changed);
	_question_n = 0;
	if (   _area_output
	    && _area_output->visible()) {
		_area_output->hide();
	} else if (   _area_output_overlay
	           && _area_output_overlay->visible()) {
		_area_output_overlay->hide();
	}
}



unsigned int
Control::answers_cnt()
{
	return _trivias[_question_n].answers().size();
}



unsigned int
Control::correct()
{
	return _correct_pos;
}



bool
Control::answer_normal(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(normal);

		return true;
	}

	return false;
}



bool
Control::answer_selected(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(selected);
		_configuration->audio()->play(Audio::answer_select);

		return true;
	}

	return false;
}



bool
Control::answer_all_correct()
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}

	free_canvas();
	std::vector<ItemAnswer*> answers =
		_configuration->item_top()->answers();
	for (std::vector<ItemAnswer*>::iterator i = answers.begin();
	     i < answers.end(); i++) {
		if ((*i)->state() == selected) {
			(*i)->mark_state(active);
		}
	}

	if (! _trivias[_question_n].answer_image().empty()) {
		image_overlay_prepare(_trivias[_question_n].answer_image());
	}

	play_answer_audio(Audio::answer_correct);

	return true;
}



bool
Control::answer_correct_correct()
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}

	free_canvas();
	std::vector<ItemAnswer*> answers = _configuration->item_top()->answers();
	if (correct() < answers.size()) {
		if (answers[correct()]->state() == selected) {
			answers[correct()]->mark_state(active);
			play_answer_audio(Audio::answer_correct);

			if (! _trivias[_question_n].answer_image().empty()) {
				image_overlay_prepare(_trivias[_question_n].answer_image());
			}

			return true;
		} else {
			answers[correct()]->mark_state(prelight);
			for (std::vector<ItemAnswer*>::iterator i = answers.begin();
			     i < answers.end(); i++) {
				if ((*i)->state() == selected) {
					(*i)->mark_state(inactive);
				}
			}
			_configuration->audio()->play(Audio::answer_wrong);

			if (! _trivias[_question_n].answer_image().empty()) {
				image_overlay_prepare(_trivias[_question_n].answer_image());
			}

			return false;
		}
	}

	throw ParserError(_("correct answer does not exist"));
}



bool
Control::answer_correct(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(active);
		play_answer_audio(Audio::answer_correct);

		return true;
	}

	return false;
}



bool
Control::answer_wrong(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(inactive);

		return true;
	}

	return false;
}



bool
Control::answer_corrected(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(prelight);

		return true;
	}

	return false;
}



bool
Control::answer_insensitive(unsigned int n)
{
	if (   _state < STATE_question
	    || _state > STATE_show_answer) {
		return false;
	}
	const std::vector<ItemAnswer*>& answers =
		_configuration->item_top()->answers();
	if (n < answers.size()) {
		free_canvas();

		answers[n]->mark_state(insensitive);

		return true;
	}

	return false;
}



void
Control::play_answer_audio(Audio::audio_t audio)
{
	if (_trivias[_question_n].answer_audio().empty()) {
		_configuration->audio()->play(audio);
	} else {
		Audio::play(_trivias[_question_n].answer_audio());
	}
}



bool
Control::prev_level()
{
	if (   _state >= STATE_question
	    && _state <= STATE_show_answer
	    && _level_n > 0) {
		free_canvas();

		if (_level_n < _level_cnt) {
			_configuration->item_top()->mark_state_level(_level_n, normal);
		}
		_level_n--;
		_configuration->item_top()->mark_state_level(_level_n, selected);
		const std::vector<ItemStatus*>& statuss =
			_configuration->item_top()->statuss();
		for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
		     i < statuss.end(); i++) {
			(*i)->set_level(_level_n);
		}

		return true;
	}

	return false;
}



bool
Control::next_level()
{
	if (   _state >= STATE_question
	    && _state <= STATE_show_answer
	    && _level_n + 1 < _level_cnt) {
		free_canvas();

		if (_level_n >= 0) {
			_configuration->item_top()->mark_state_level(_level_n, active);
		}
		_level_n++;
		_configuration->item_top()->mark_state_level(_level_n, selected);
		const std::vector<ItemStatus*>& statuss =
			_configuration->item_top()->statuss();
		for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
		     i < statuss.end(); i++) {
			(*i)->set_level(_level_n);
			(*i)->next_level();
		}

		return true;
	}

	return false;
}



void
Control::reset_level()
{
	_configuration->item_top()->mark_state_level(0, selected);
	for (Trivias::size_type i = 1; i <= _level_n; i++) {
		_configuration->item_top()->mark_state_level(i, normal);
	}
	_level_n = 0;
	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
	     i < statuss.end(); i++) {
		(*i)->set_level(_level_n);
	}
}



void
Control::action_activate(Plugin* plugin, const Glib::ustring& arg)
{
	image_shrink();
	_action_active = plugin;
	_action_active->activate(arg);
}



void
Control::action_deactivate()
{
	image_shrink();
	if (_action_active) {
		_action_active->deactivate();
		_action_active = 0;
	}
}



void
Control::free_canvas()
{
	image_shrink();
	if (_action_active) {
		_action_active->free_canvas();
		_action_active = 0;
	}
}



void
Control::next_status(unsigned int number)
{
	free_canvas();
	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
	     i < statuss.end(); i++) {
		(*i)->next_manual(number);
	}
}



void
Control::prev_status(unsigned int number)
{
	free_canvas();
	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
	     i < statuss.end(); i++) {
		(*i)->prev_manual(number);
	}
}



void
Control::set_status(unsigned int number, unsigned int n)
{
	free_canvas();
	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
	     i < statuss.end(); i++) {
		(*i)->set_manual(number, n);
	}
}



void
Control::reset_status()
{
	const std::vector<ItemStatus*>& statuss =
		_configuration->item_top()->statuss();
	for (std::vector<ItemStatus*>::const_iterator i = statuss.begin();
	     i < statuss.end(); i++) {
		(*i)->reset();
	}
}





} // end namespace quimeleon
