/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2006  Aike Reyer <aike@users.sourceforge.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "data_dirs_editor.h"

#include <gtkmm/button.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>

#include "gettext.h"


#include <iostream>





namespace quimeleon {



DataDirsEditor::DataDirsEditor(Configuration* configuration)
	: _configuration(configuration),
	  _selection(_list_view_text.get_selection()),
	  _button_add(Gtk::Stock::ADD),
	  _button_remove(Gtk::Stock::REMOVE),
	  _button_browse(_8("_Browse ..."), true),
	  _button_up(Gtk::Stock::GO_UP),
	  _button_down(Gtk::Stock::GO_DOWN)

{
	const std::vector<std::string>& data_dirs = _configuration->data_dirs();
	int i;
	for (i = data_dirs.size() - 1;
	     i >= int(_configuration->global_data_dirs_size()); i--) {
		_list_view_text.append_text(data_dirs[i]);
	}
	for (; i >= 0; i--) {
		_list_view_text.append_disabled_text(data_dirs[i]);
	}


	_button_box.set_layout(Gtk::BUTTONBOX_START);

	_button_add.set_alignment(0.0, 0.5);
	_button_add.signal_clicked().connect(
		sigc::mem_fun(*this, &DataDirsEditor::add_dir));
	_button_box.pack_start(_button_add, Gtk::PACK_SHRINK);

	_button_remove.set_alignment(0.0, 0.5);
	_button_remove.signal_clicked().connect(
		sigc::mem_fun(*this, &DataDirsEditor::remove_dir));
	_button_box.pack_start(_button_remove, Gtk::PACK_SHRINK);

	_button_browse.signal_clicked().connect(
		sigc::mem_fun(*this, &DataDirsEditor::browse_dir));
	_button_box.pack_start(_button_browse, Gtk::PACK_SHRINK);

	_button_up.set_alignment(0.0, 0.5);
	_button_up.signal_clicked().connect(
		sigc::mem_fun(*this, &DataDirsEditor::move_up_dir));
	_button_box.pack_end(_button_up, Gtk::PACK_SHRINK);
	_button_box.set_child_secondary(_button_up);

	_button_down.set_alignment(0.0, 0.5);
	_button_down.signal_clicked().connect(
		sigc::mem_fun(*this, &DataDirsEditor::move_down_dir));
	_button_box.pack_end(_button_down, Gtk::PACK_SHRINK);
	_button_box.set_child_secondary(_button_down);

	enable_edit(false);

	_list_view_text.signal_edit_enabled().connect(
		sigc::mem_fun(*this, &DataDirsEditor::enable_edit));

	show_all();
}



Glib::ustring
DataDirsEditor::select_folder(const Glib::ustring& dir)
{
	Gtk::FileChooserDialog file_chooser_dialog(
		_8("Select folder"),
		Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER);
	file_chooser_dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	file_chooser_dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (! dir.empty()) {
		file_chooser_dialog.set_filename(absolute_path(dir));
	}


	if (file_chooser_dialog.run() == Gtk::RESPONSE_OK) {
		return file_chooser_dialog.get_filename();
	} else {
		return "";
	}
}



void
DataDirsEditor::add_dir()
{
	Glib::ustring folder = select_folder(_list_view_text.get_selected_text());

	if (! folder.empty()) {
		_list_view_text.insert_text(folder);
	}
}



void
DataDirsEditor::remove_dir()
{
	_list_view_text.erase_selected();
}



void
DataDirsEditor::browse_dir()
{
	Glib::ustring folder = select_folder(_list_view_text.get_selected_text());

	if (! folder.empty()) {
		_list_view_text.set_selected_text(folder);
	}
}



void
DataDirsEditor::move_up_dir()
{
	_list_view_text.move_selected_up();
}



void
DataDirsEditor::move_down_dir()
{
	_list_view_text.move_selected_down();
}



void
DataDirsEditor::enable_edit(bool do_enable)
{
	_button_remove.set_sensitive(do_enable);
	_button_browse.set_sensitive(do_enable);
	_button_up.set_sensitive(do_enable);
	_button_down.set_sensitive(do_enable);
}



std::vector<std::string>
DataDirsEditor::data_dirs() const
{
	std::vector<std::string> re;

	Glib::RefPtr<const Gtk::TreeModel> model = _list_view_text.get_model();

	Gtk::TreeModel::Children children = model->children();


/*	// crashes under weird circumstances (DATADIRROOT does not exist)
	for (Gtk::TreeModel::Children::reverse_iterator i = children.rbegin();
	     i != children.rend(); i++) {
		Glib::ustring dir = (*i)[_list_view_text.column_record()._text];
		if (G_IS_DIR_SEPARATOR(dir[dir.length() - 1])) {
			re.push_back(dir.substr(0, dir.length() - 1));
		} else {
			re.push_back(dir);
		}
	}
*/
	for (int i = children.size() - 1; i >= 0; i--) {
		Glib::ustring dir = children[i][_list_view_text.column_record()._text];
		if (G_IS_DIR_SEPARATOR(dir[dir.length() - 1])) {
			re.push_back(dir.substr(0, dir.length() - 1));
		} else {
			re.push_back(dir);
		}
	}


	return re;
}





} // end namespace quimeleon

