/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include <iomanip>

#include "audio.h"
#include "item.h"
#include "item_action.h"
#include "item_answer.h"
#include "item_area.h"
#include "item_level.h"
#include "item_question.h"




namespace quimeleon {





Glib::RefPtr<Gdk::Pixbuf> Item::_pixbuf;
Glib::RefPtr<Gdk::GC> Item::_gc;
Glib::RefPtr<Gdk::Drawable> Item::_drawable;
Glib::RefPtr<Pango::Context> Item::_pango_context;
float Item::_font_scale = 1.0;





Item::Item(const Item* parent, const Glib::ustring& id, item_type_t type)
	: _parse_alternatives(false),
	  _parent(parent),
	  _x(0), _x_is_percent(false),
	  _y(0), _y_is_percent(false),
	  _width(100), _width_is_percent(true),
	  _height(100), _height_is_percent(true),
	  _width_is_set(false),
	  _height_is_set(false),
	  _anchor(NW),
	  _alpha_normal(0xff),
	  _alpha_selected(-1),
	  _alpha_active(-1),
	  _alpha_prelight(-1),
	  _alpha_inactive(-1),
	  _alpha_insensitive(-1),
	  _color_normal_r(127),
	  _color_normal_g(127),
	  _color_normal_b(127),
	  _color_selected_r(-1),
	  _color_selected_g(-1),
	  _color_selected_b(-1),
	  _color_active_r(-1),
	  _color_active_g(-1),
	  _color_active_b(-1),
	  _color_prelight_r(-1),
	  _color_prelight_g(-1),
	  _color_prelight_b(-1),
	  _color_inactive_r(-1),
	  _color_inactive_g(-1),
	  _color_inactive_b(-1),
	  _color_insensitive_r(-1),
	  _color_insensitive_g(-1),
	  _color_insensitive_b(-1),
	  _id(id),
	  _type(type),
	  _silence(false),
	  _alternative(0),
	  _preferred_type(AT_both),
	  _state(normal),
	  _is_visible(true),
	  _question(0)
{
}



Item::~Item()
{
	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
	     j < i->end(); j++) {
			delete *j;
		}
	}
	if (_connection_expose) {
		_connection_expose.disconnect();
	}
}




void
Item::reset()
{
	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
	     j < i->end(); j++) {
			delete *j;
		}
	}

	_children.clear();
	_answers.clear();
	_question = 0;
	_actions.clear();
	_levels.clear();
	_statuss.clear();
	_areas.clear();


	_alpha_normal = 0xff;
	_color_normal_r = 127;
	_color_normal_g = 127;
	_color_normal_b = 127;

	_alpha_selected = -1;
	_alpha_active = -1;
	_alpha_prelight = -1;
	_alpha_inactive = -1;
	_alpha_insensitive = -1;
	_color_selected_r = -1;
	_color_selected_g = -1;
	_color_selected_b = -1;
	_color_active_r = -1;
	_color_active_g = -1;
	_color_active_b = -1;
	_color_prelight_r = -1;
	_color_prelight_g = -1;
	_color_prelight_b = -1;
	_color_inactive_r = -1;
	_color_inactive_g = -1;
	_color_inactive_b = -1;
	_color_insensitive_r = -1;
	_color_insensitive_g = -1;
	_color_insensitive_b = -1;

	_parse_alternatives = false;
	_x = 0;
	_x_is_percent = false;
	_y = 0;
	_y_is_percent = false;
	_width = 100;
	_width_is_percent = true;
	_height = 100;
	_height_is_percent = true;
	_width_is_set = false;
	_height_is_set = false;
	_anchor = NW;
	_silence = false;
	_alternative = 0;
	_preferred_type = AT_both;
	_state = normal;
	_is_visible = true;


	_pixbuf.clear();
}



void
Item::parse_alternatives(bool parse_alternatives)
{
	_parse_alternatives = parse_alternatives;


	if (_parse_alternatives) {
		_children.push_back(std::vector<Item*>());
	} else if (_children.back().empty()) {	// implies ! _children.empty()
		_children.pop_back();
	} else {	// implies ! _children.empty()
		std::vector<Item*>& alternatives = _children.back();
		std::sort(alternatives.begin(), alternatives.end(), cmp_answers());
	}
}



void
Item::complete_attributes()
{
	if (_alpha_selected < 0) { _alpha_selected = _alpha_normal; }
	if (_alpha_active < 0) { _alpha_active = _alpha_normal; }
	if (_alpha_prelight < 0) { _alpha_prelight = _alpha_normal; }
	if (_alpha_inactive < 0) { _alpha_inactive = _alpha_normal; }
	if (_alpha_insensitive < 0) { _alpha_insensitive = _alpha_normal; }
	if (_color_selected_r < 0) { _color_selected_r = _color_normal_r; }
	if (_color_selected_g < 0) { _color_selected_g = _color_normal_g; }
	if (_color_selected_b < 0) { _color_selected_b = _color_normal_b; }
	if (_color_active_r < 0) { _color_active_r = _color_normal_r; }
	if (_color_active_g < 0) { _color_active_g = _color_normal_g; }
	if (_color_active_b < 0) { _color_active_b = _color_normal_b; }
	if (_color_prelight_r < 0) { _color_prelight_r = _color_normal_r; }
	if (_color_prelight_g < 0) { _color_prelight_g = _color_normal_g; }
	if (_color_prelight_b < 0) { _color_prelight_b = _color_normal_b; }
	if (_color_inactive_r < 0) { _color_inactive_r = _color_normal_r; }
	if (_color_inactive_g < 0) { _color_inactive_g = _color_normal_g; }
	if (_color_inactive_b < 0) { _color_inactive_b = _color_normal_b; }
	if (_color_insensitive_r < 0) { _color_insensitive_r = _color_normal_r; }
	if (_color_insensitive_g < 0) { _color_insensitive_g = _color_normal_g; }
	if (_color_insensitive_b < 0) { _color_insensitive_b = _color_normal_b; }
}



gushort
Item::r() const throw()
{
	switch (_state) {
	case selected:
		return _color_selected_r;
	case active:
		return _color_active_r;
	case prelight:
		return _color_prelight_r;
	case inactive:
		return _color_inactive_r;
	case insensitive:
		return _color_insensitive_r;
	case normal:
	default:
		return _color_normal_r;
	}
}



gushort
Item::g() const throw()
{
	switch (_state) {
	case selected:
		return _color_selected_g;
	case active:
		return _color_active_g;
	case prelight:
		return _color_prelight_g;
	case inactive:
		return _color_inactive_g;
	case insensitive:
		return _color_insensitive_g;
	case normal:
	default:
		return _color_normal_g;
	}
}



gushort
Item::b() const throw()
{
	switch (_state) {
	case selected:
		return _color_selected_b;
	case active:
		return _color_active_b;
	case prelight:
		return _color_prelight_b;
	case inactive:
		return _color_inactive_b;
	case insensitive:
		return _color_insensitive_b;
	case normal:
	default:
		return _color_normal_b;
	}
}



int
Item::alpha() const throw()
{
	switch (_state) {
	case selected:
		return _alpha_selected;
	case active:
		return _alpha_active;
	case prelight:
		return _alpha_prelight;
	case inactive:
		return _alpha_inactive;
	case insensitive:
		return _alpha_insensitive;
	case normal:
	default:
		return _alpha_normal;
	}
}



bool
Item::set_alternative(int n, answer_type_t preferred_type)
{
	if (   _alternative != n
	    || _preferred_type != preferred_type) {
		for (std::vector<std::vector<Item*> >::const_iterator i
		     	= _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {
				(*j)->set_alternative(n, preferred_type);
			}
		}

		_alternative = n;
		_preferred_type = preferred_type;

		_answers.clear();
		_question = 0;
		_actions.clear();
		_levels.clear();
		_statuss.clear();
		_areas.clear();

		return true;
	}

	return false;
}



void
Item::set_parent_dimensions(int x, int y, int width, int height,
	int crop_x, int crop_y, int crop_width, int crop_height)
{
	_parent_x = x;
	_parent_y = y;
	_parent_width = width;
	_parent_height = height;

	if (_width_is_percent) {
		_real_width = int(_width * _parent_width) / 100;
	} else {
		_real_width = int(_width);
	}
	if (_height_is_percent) {
		_real_height = int(_height * _parent_height) / 100;
	} else {
		_real_height = int(_height);
	}

	_real_x = _parent_x + int(_x_is_percent ? _parent_width * _x / 100 : _x);
	switch (_anchor) {
	case N:
	case C:
	case S:
		_real_x -= _real_width / 2;
		break;
	case NE:
	case E:
	case SE:
		_real_x -= _real_width;
		break;
	default: {}
	}

	_real_y = _parent_y + int(_y_is_percent ? _parent_height * _y / 100 : _y);
	switch (_anchor) {
	case W:
	case C:
	case E:
		_real_y -= _real_height / 2;
		break;
	case SW:
	case S:
	case SE:
		_real_y -= _real_height;
		break;
	default: {}
	}


	if (_real_x < crop_x) {
		_crop_x = crop_x;
	} else {
		_crop_x = _real_x;
	}
	const int real_x2 = _real_x + _real_width;
	int crop_x2 = crop_x + crop_width;
	if (real_x2 < crop_x2) {
		crop_x2 = real_x2;
	}
	_crop_width = crop_x2 - _crop_x;

	if (_real_y < crop_y) {
		_crop_y = crop_y;
	} else {
		_crop_y = _real_y;
	}
	const int real_y2 = _real_y + _real_height;
	int crop_y2 = crop_y + crop_height;
	if (real_y2 < crop_y2) {
		crop_y2 = real_y2;
	}
	_crop_height = crop_y2 - _crop_y;



	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_parent_dimensions(
				_real_x, _real_y, _real_width, _real_height,
				_crop_x, _crop_y, _crop_width, _crop_height);
		}
	}
}



void
Item::set_state_rec(state_t state)
{
	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_state_rec(state);
		}
	}
	set_state(state);
}



void
Item::connect_signal_expose(const sigc::slot<void, int, int, int, int>& slot)
{
	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->connect_signal_expose(slot);
		}
	}

	if (_connection_expose) {
		_connection_expose.disconnect();
	}
	_connection_expose = _signal_expose.connect(slot);
}



void
Item::signal_expose() const
{
	if (   crop_width() > 0
	    && crop_height() > 0) {
		_signal_expose.emit(crop_x(), crop_y(), crop_width(), crop_height());
	}
}



ItemQuestion*
Item::question() const
{
// _question is set to 0 in Item::set_alternative()
	if (! _question) {
		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					_question = (*j)->question();

					if (_question) {
						return _question;
					}
				}
			}
		}

	}

	return _question;
}



const std::vector<ItemAnswer*>&
Item::answers() const
{
// _answers is clear()ed in Item::set_alternative()
	if (_answers.empty()) {
		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {

					std::copy(child_answers.begin(), child_answers.end(),
						std::back_inserter(_answers));

					break;
				}
			}
		}


		std::sort(_answers.begin(), _answers.end(), ItemAnswer::cmp_id());

	}

	return _answers;
}



const std::vector<ItemAction*>&
Item::actions() const
{
// _actions is clear()ed in Item::set_alternative()
	if (_actions.empty()) {

		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					const std::vector<ItemAction*>& child_actions
						= (*j)->actions();

					std::copy(child_actions.begin(), child_actions.end(),
						std::back_inserter(_actions));

					break;
				}
			}
		}


		std::sort(_actions.begin(), _actions.end(), ItemAction::cmp_id());
	}

	return _actions;
}



const std::vector<ItemAction*>
Item::all_actions() const
{
	std::vector<ItemAction*> all_actions;

	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {

			const std::vector<ItemAction*>& child_all_actions
				= (*j)->all_actions();

			std::copy(child_all_actions.begin(), child_all_actions.end(),
				std::back_inserter(all_actions));
		}
	}


	return all_actions;
}



const std::vector<ItemLevel*>&
Item::levels() const
{
// _levels is clear()ed in Item::set_alternative()
	if (_levels.empty()) {

		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					const std::vector<ItemLevel*>& child_levels
						= (*j)->levels();

					std::copy(child_levels.begin(), child_levels.end(),
						std::back_inserter(_levels));

					break;
				}
			}
		}


		std::sort(_levels.begin(), _levels.end(), ItemLevel::cmp_id());
	}

	return _levels;
}



const std::vector<ItemLevel*>
Item::all_levels() const
{
	std::vector<ItemLevel*> all_levels;

	for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {

			const std::vector<ItemLevel*>& child_all_levels
				= (*j)->all_levels();

			std::copy(child_all_levels.begin(), child_all_levels.end(),
				std::back_inserter(all_levels));
		}
	}

	std::sort(all_levels.begin(), all_levels.end(), ItemLevel::cmp_id());

	return all_levels;
}



const std::vector<ItemStatus*>&
Item::statuss() const
{
// _statuss is clear()ed in Item::set_alternative()
	if (_statuss.empty()) {

		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					const std::vector<ItemStatus*>& child_statuss
						= (*j)->statuss();

					std::copy(child_statuss.begin(), child_statuss.end(),
						std::back_inserter(_statuss));

					break;
				}
			}
		}
	}

	return _statuss;
}



const std::map<const Glib::ustring, ItemArea*>&
Item::areas() const
{
// _areas is clear()ed in Item::set_alternative()
	if (_areas.empty()) {

		for (std::vector<std::vector<Item*> >::const_iterator i = _children.begin();
		     i < _children.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {

				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					const std::map<const Glib::ustring, ItemArea*>& child_areas
						= (*j)->areas();

					for (std::map<const Glib::ustring,
					              ItemArea*>::const_iterator i
					     	= child_areas.begin();
					     i != child_areas.end(); i++) {
						_areas.insert(std::pair<const Glib::ustring, ItemArea*>(
							i->first, i->second));
					}

					break;
				}
			}
		}
	}

	return _areas;
}




Item*
Item::find_id(const Glib::ustring& id)
{
	if (id == _id) {
		return this;
	}

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			Item* re = (*j)->find_id(id);
			if (re) {
				return re;
			}
		}
	}


	return 0;
}



void
Item::compose_pixbuf()
{
	if (! visible()) {
		return;
	}

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= children().begin();
	     i < children().end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {
				const std::vector<ItemAnswer*>& child_answers = (*j)->answers();

// if there is enough room for the answers in the child: compose_pixbuf()
// and break
				if (   int(child_answers.size()) >= _alternative
				    || j == i->end() - 1) {
					(*j)->compose_pixbuf();

					break;
				}
			}
	}

	 if (silence()) {
		 Audio::mute();
	 } else if (! soundfile().empty()) {
		 Audio::play(soundfile());
	 }
}



Item*
Item::find_type(item_type_t type)
{
	if (_type == type) {
		return this;
	}

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children.begin();
	     i < _children.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			Item* re = (*j)->find_type(type);
			if (re) {
				return re;
			}
		}
	}

	return 0;
}





} // end namespace quimeleon
