/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifndef QUIMELEON_ITEM_H
#define QUIMELEON_ITEM_H

#include "error.h"

#include <iostream>
#include <sstream>
#include <vector>

#include <glibmm/ustring.h>
#include <gtkmm/drawingarea.h>




namespace quimeleon {





enum anchor_t {
	N,
	NE,
	E,
	SE,
	S,
	SW,
	W,
	NW,

	C
};



enum item_type_t {
	T_unknown = 0,
	T_rect,
	T_pixmap,
	T_varpixmap,
	T_label,
	T_varlabel,
	T_area,
	T_draw,
	T_status
};



enum state_t {
	normal,
	selected,
	active,
	inactive,
	insensitive,
	prelight
};



enum id_t {
	id_unknown = 0,
	id_answer
};



enum answer_type_t {
	AT_both = 0,
	AT_text,
	AT_image
};




class ItemLabel;
class ItemAnswer;
class ItemAction;
class ItemLevel;
class ItemQuestion;
class ItemStatus;
class ItemArea;
class ItemPixbuf;





class Item
{
public:
	Item(const Item* parent = 0,
		const Glib::ustring& id = "", item_type_t type = T_unknown);
	virtual ~Item();


	virtual void parse_alternatives(bool parse_alternatives);
	virtual void complete_attributes();


	inline void set_x(float x, bool is_percent) throw()
	{
		_x = x;
		_x_is_percent = is_percent;
	}

	inline void set_y(float y, bool is_percent) throw()
	{
		_y = y;
		_y_is_percent = is_percent;
	}

	inline void set_width(float width, bool is_percent, bool is_set = true)
		throw()
	{
		_width = width;
		_width_is_percent = is_percent;
		_width_is_set = is_set;
	}
	inline float width_percent() const throw()
	{
		return (_width_is_percent && _width_is_set ? _width : 0);
	}

	inline void set_height(float height, bool is_percent, bool is_set = true)
		throw()
	{
		_height = height;
		_height_is_percent = is_percent;
		_height_is_set = is_set;
	}
	inline float height_percent() const throw()
	{
		return (_height_is_percent && _height_is_set ? _height : 0);
	}

	inline void set_anchor(anchor_t anchor) throw()
	{
		_anchor = anchor;
	}
	inline anchor_t anchor() const throw() { return _anchor; }

	inline void set_alpha_normal(int alpha) throw()
	{
		_alpha_normal = alpha;
	}

	inline void set_alpha_selected(int alpha) throw()
	{
		_alpha_selected = alpha;
	}

	inline void set_alpha_active(int alpha) throw()
	{
		_alpha_active = alpha;
	}

	inline void set_alpha_prelight(int alpha) throw()
	{
		_alpha_prelight = alpha;
	}

	inline void set_alpha_inactive(int alpha) throw()
	{
		_alpha_inactive = alpha;
	}

	inline void set_alpha_insensitive(int alpha) throw()
	{
		_alpha_insensitive = alpha;
	}


	inline void set_color_normal(int r, int g, int b) throw()
	{
		_color_normal_r = r;
		_color_normal_g = g;
		_color_normal_b = b;
	}

	inline void set_color_selected(int r, int g, int b) throw()
	{
		_color_selected_r = r;
		_color_selected_g = g;
		_color_selected_b = b;
	}

	inline void set_color_active(int r, int g, int b) throw()
	{
		_color_active_r = r;
		_color_active_g = g;
		_color_active_b = b;
	}

	inline void set_color_prelight(int r, int g, int b) throw()
	{
		_color_prelight_r = r;
		_color_prelight_g = g;
		_color_prelight_b = b;
	}

	inline void set_color_inactive(int r, int g, int b) throw()
	{
		_color_inactive_r = r;
		_color_inactive_g = g;
		_color_inactive_b = b;
	}

	inline void set_color_insensitive(int r, int g, int b) throw()
	{
		_color_insensitive_r = r;
		_color_insensitive_g = g;
		_color_insensitive_b = b;
	}


	inline void set_content(Glib::ustring content)
	{
		_content = content;
	}

	inline void set_soundfile(const Glib::ustring& file)
	{
		_soundfile = file;
	}
	inline void set_silence(bool silence)
	{
		_silence = silence;
	}
	inline const Glib::ustring& soundfile() const throw() { return _soundfile; }
	inline bool silence() const throw() { return _silence; }


	inline virtual void add_child(Item* child)
	{
		if (_parse_alternatives) {
			_children.back().push_back(child);
		} else {
			_children.push_back(std::vector<Item*>(1, child));
		}
	}




	gushort r() const throw();
	gushort g() const throw();
	gushort b() const throw();
	int alpha() const throw();

	inline const Glib::ustring& id() const throw() { return _id; }
	const Glib::ustring& content() const throw() { return _content; }

	inline int x() const throw() { return _real_x; }
	inline int y() const throw() { return _real_y; }
	inline int width() const throw() { return _real_width; }
	inline int height() const throw() { return _real_height; }

	inline int crop_x() const throw() { return _crop_x; }
	inline int crop_y() const throw() { return _crop_y; }
	inline int crop_width() const throw() { return _crop_width; }
	inline int crop_height() const throw() { return _crop_height; }

	inline const std::vector<std::vector<Item*> >& children() const throw()
	{
		return _children;
	}



	bool set_alternative(int n, answer_type_t preferred_type);
	virtual void set_parent_dimensions(int x, int y, int width, int height,
		int crop_x, int crop_y, int crop_width, int crop_height);
	inline void set_state(state_t state) { _state = state; }
	void set_state_rec(state_t state);
	inline void set_visible(bool is_visible = true) throw()
	{
		_is_visible = is_visible;
	}

	inline state_t state() const throw() { return _state; }
	inline bool visible() const throw() { return _is_visible; };


	void connect_signal_expose(
		const sigc::slot<void, int, int, int, int>& slot);
	void signal_expose() const;

	virtual ItemQuestion* question() const;
	virtual const std::vector<ItemAnswer*>& answers() const;
	virtual const std::vector<ItemAction*>& actions() const;
	virtual const std::vector<ItemAction*> all_actions() const;
	virtual const std::vector<ItemLevel*>& levels() const;
	virtual const std::vector<ItemLevel*> all_levels() const;
	virtual const std::vector<ItemStatus*>& statuss() const;
	virtual const std::map<const Glib::ustring, ItemArea*>& areas() const;

	Item* find_id(const Glib::ustring& id);

	virtual void compose_pixbuf();


protected:
	virtual void reset();

	inline bool parse_alternatives() const throw()
	{
		return _parse_alternatives;
	}

	inline int alternative() const throw() { return _alternative; }
	inline answer_type_t preferred_type() const throw()
	{
		return _preferred_type;
	}

	inline int parent_x() const throw() { return _parent_x; }
	inline int parent_y() const throw() { return _parent_y; }
	inline int parent_width() const throw() { return _parent_width; }
	inline int parent_height() const throw() { return _parent_height; }

	Item* find_type(item_type_t type);


public:
	static inline void
	set_drawable(Glib::RefPtr<Gdk::Drawable> drawable)
	{
		_gc = Gdk::GC::create(drawable);
		_drawable = drawable;
	}
	static inline const Glib::RefPtr<Gdk::Drawable>& get_drawable()
	{
		return _drawable;
	}

	static inline Glib::RefPtr<Gdk::GC> gc()
	{
		return _gc;
	}

	static inline void
	set_pango_context(Glib::RefPtr<Pango::Context> pango_context)
	{
		_pango_context = pango_context;
	}

	static inline Glib::RefPtr<Pango::Context> pango_context()
	{
		return _pango_context;
	}
	static inline float font_scale() throw() { return _font_scale; }

protected:
	static inline void set_font_scale(int width, int height) throw()
	{
		_font_scale = float(width) / 800.0;
	}


protected:
	static Glib::RefPtr<Gdk::Pixbuf> _pixbuf;
	static Glib::RefPtr<Gdk::GC> _gc;
	static Glib::RefPtr<Gdk::Drawable> _drawable;
	static Glib::RefPtr<Pango::Context> _pango_context;


private:
	bool _parse_alternatives;

	const Item* _parent;
	std::vector<std::vector<Item*> > _children;

	float _x;
	bool _x_is_percent;
	float _y;
	bool _y_is_percent;
	float _width;
	bool _width_is_percent;
	float _height;
	bool _height_is_percent;
	bool _width_is_set;
	bool _height_is_set;
	anchor_t _anchor;

protected:
	int _alpha_normal;
	int _alpha_selected;
	int _alpha_active;
	int _alpha_prelight;
	int _alpha_inactive;
	int _alpha_insensitive;
	int _color_normal_r;
	int _color_normal_g;
	int _color_normal_b;
	int _color_selected_r;
	int _color_selected_g;
	int _color_selected_b;
	int _color_active_r;
	int _color_active_g;
	int _color_active_b;
	int _color_prelight_r;
	int _color_prelight_g;
	int _color_prelight_b;
	int _color_inactive_r;
	int _color_inactive_g;
	int _color_inactive_b;
	int _color_insensitive_r;
	int _color_insensitive_g;
	int _color_insensitive_b;

private:
	const Glib::ustring _id;
	const item_type_t _type;

	Glib::ustring _content;
	Glib::ustring _soundfile;
	bool _silence;


	int _parent_width;
	int _parent_height;
	int _parent_x;
	int _parent_y;
	int _real_x;
	int _real_y;
	int _real_width;
	int _real_height;
	int _crop_x;
	int _crop_y;
	int _crop_width;
	int _crop_height;

	int _alternative;
	answer_type_t _preferred_type;
	state_t _state;
	bool _is_visible;

	sigc::signal<void, int, int, int, int> _signal_expose;
	sigc::connection _connection_expose;


protected:
	mutable ItemQuestion* _question;
	mutable std::vector<ItemAnswer*> _answers;
	mutable std::vector<ItemAction*> _actions;
	mutable std::vector<ItemLevel*> _levels;
	mutable std::vector<ItemStatus*> _statuss;
	mutable std::map<const Glib::ustring, ItemArea*> _areas;

private:
	class cmp_answers
	{
	public:
		inline int operator() (Item const* a, Item const* b) const
		{
			return a->answers().size() < b->answers().size();
		}
	};

private:
	static float _font_scale;
};





} // end namespace quimeleon




#endif
