/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "item_pixmap.h"

#include "gettext.h"





namespace quimeleon {





std::map<const Glib::ustring, std::map<std::pair<int, int>, Glib::RefPtr<Gdk::Pixbuf> > > ItemPixmap::_pixbufs;



ItemPixmap::ItemPixmap(const Item* parent,
	const Glib::ustring& id, item_type_t type,
	bool keep_aspect,
	bool do_buffer)
	: Item(parent, id, type),
	  _keep_aspect(keep_aspect),
	  _do_buffer(do_buffer),
	  _pixbuf_w(0),
	  _pixbuf_h(0)
{
}



void
ItemPixmap::complete_attributes()
{
	if (_selected.empty()) { _selected = _normal; }
	if (_active.empty()) { _active = _normal; }
	if (_prelight.empty()) { _prelight = _normal; }
	if (_inactive.empty()) { _inactive = _normal; }
	if (_insensitive.empty()) { _insensitive = _normal; }

	Item::complete_attributes();
}



void
ItemPixmap::set_parent_dimensions(int x, int y, int width, int height,
	int crop_x, int crop_y, int crop_width, int crop_height)
{
	_pixbuf_w = 0;
	_pixbuf_h = 0;

	Item::set_parent_dimensions(x, y, width, height,
		crop_x, crop_y, crop_width, crop_height);
}



void
ItemPixmap::compose_pixbuf()
{
	if (   ! visible()
	    || ! _pixbuf) {
		return;
	}

	if (alpha() > 0) {
		try {
			Glib::ustring filename;

			switch (state()) {
			case selected:
				filename = _selected;
				break;
			case active:
				filename = _active;
				break;
			case prelight:
				filename = _prelight;
				break;
			case inactive:
				filename = _inactive;
				break;
			case insensitive:
				filename = _insensitive;
				break;
			default:
				filename = _normal;
			}

			Glib::RefPtr<Gdk::Pixbuf> pixbuf;
			std::map<const Glib::ustring,
			         std::map<std::pair<int, int>,
			                  Glib::RefPtr<Gdk::Pixbuf> > >::const_iterator
				pixbuf_filename;
			std::map<std::pair<int, int>,
			         Glib::RefPtr<Gdk::Pixbuf> >::const_iterator
				pixbuf_filename_size;
			if (   _do_buffer
			    && _pixbuf_w
			    && _pixbuf_h
			    && (pixbuf_filename = _pixbufs.find(filename))
				    	!= _pixbufs.end()
				&& (pixbuf_filename_size = pixbuf_filename->second.find(
				    	std::pair<int, int>(_pixbuf_w, _pixbuf_h)))
				    		!= pixbuf_filename->second.end()) {
				pixbuf = pixbuf_filename_size->second;
			} else {
				try {
					pixbuf = Gdk::Pixbuf::create_from_file(filename);
				} catch (const Gdk::PixbufError& e) {
					if (filename != _normal) {
						std::cerr << _("Could not load image: ") << "\n  "
							<< e.what() << '\n';
						if (e.code() == Gdk::PixbufError::UNKNOWN_TYPE) {
							std::cerr << _("This image file format is not supported by your version of GTK+/GdkPixbuf.\nThe command gdk-pixbuf-query-loaders lists supported formats.") << "\n\n";
						}

						pixbuf = Gdk::Pixbuf::create_from_file(_normal);
					} else {
						throw;
					}
				} catch (const Glib::Exception& e) {
					if (filename != _normal) {
						std::cerr << e.what() << '\n';

						pixbuf = Gdk::Pixbuf::create_from_file(_normal);
					} else {
						throw;
					}
				}

				_pixbuf_w = width();

				if (_keep_aspect) {
					_pixbuf_h =
						_pixbuf_w * pixbuf->get_height() / pixbuf->get_width();

					if (_pixbuf_h > height()) {
						_pixbuf_h = height();
						_pixbuf_w =
							_pixbuf_h * pixbuf->get_width() / pixbuf->get_height();
					}

				} else {
					_pixbuf_h = height();
				}

				if (   (pixbuf_filename = _pixbufs.find(filename))
					    	!= _pixbufs.end()
				    && (pixbuf_filename_size = pixbuf_filename->second.find(
					    	std::pair<int, int>(_pixbuf_w, _pixbuf_h)))
					    		!= pixbuf_filename->second.end()) {
					pixbuf = pixbuf_filename_size->second;
				} else if (   _pixbuf_w > 0
				           && _pixbuf_h > 0) {
					pixbuf = pixbuf->scale_simple(_pixbuf_w, _pixbuf_h,
					                              Gdk::INTERP_BILINEAR);
					if (_do_buffer) {
						_pixbufs[filename][std::pair<int, int>(_pixbuf_w, _pixbuf_h)]
							= pixbuf;
					}
				}
			}

			int dest_x = x();
			switch (anchor()) {
			case N:
			case C:
			case S:
				dest_x += (width() - _pixbuf_w) / 2;
				break;
			case NE:
			case E:
			case SE:
				dest_x += width() - _pixbuf_w;
				break;
			default: {}
			}

			int dest_y = y();
			switch (anchor()) {
			case W:
			case C:
			case E:
				dest_y += (height() - _pixbuf_h) / 2;
				break;
			case SW:
			case S:
			case SE:
				dest_y += height() - _pixbuf_h;
				break;
			default: {}
			}


			const int x2 = dest_x + _pixbuf_w;
			int crop_x2 = crop_x() + crop_width();
			if (x2 < crop_x2) {
				crop_x2 = x2;
			}
			const int y2 = dest_y + _pixbuf_h;
			int crop_y2 = crop_y() + crop_height();
			if (y2 < crop_y2) {
				crop_y2 = y2;
			}

			int crop_x1 = crop_x();
			if (dest_x > crop_x()) {
				crop_x1 = dest_x;
			}
			int crop_y1 = crop_y();
			if (dest_y > crop_y()) {
				crop_y1 = dest_y;
			}


			int crop_width = crop_x2 - crop_x1;
			int crop_height = crop_y2 - crop_y1;
			if (   dest_x != x()
			    || dest_y != y()
			    || _pixbuf_w != width()
			    || _pixbuf_h != height()) {
				for (std::vector<std::vector<Item*> >::const_iterator i
				     	= children().begin();
				     i < children().end(); i++) {
					for (std::vector<Item*>::const_iterator j = i->begin();
					     j < i->end(); j++) {
						(*j)->set_parent_dimensions(
							dest_x, dest_y, _pixbuf_w, _pixbuf_h,
							crop_x1, crop_y1, crop_width, crop_height);
					}
				}
			}

			if (   crop_width > 0
			    && crop_height > 0) {
				pixbuf->composite(_pixbuf,
					crop_x1, crop_y1, crop_width, crop_height,
					dest_x, dest_y,
					1, 1,
					Gdk::INTERP_BILINEAR, alpha());
			}
		} catch (const Gdk::PixbufError& e) {
			std::cerr << _("Could not load image: ") << e.what() << '\n';
			if (e.code() == Gdk::PixbufError::UNKNOWN_TYPE) {
				std::cerr << _("This image file format is not supported by your version of GTK+/GdkPixbuf.\nThe command gdk-pixbuf-query-loaders lists supported formats.") << "\n\n";
			}
		} catch (const Glib::Exception& e) {
			std::cerr << e.what() << '\n';
		}
	}

	Item::compose_pixbuf();
}





} // end namespace quimeleon
