/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "item_rect.h"

#include "gettext.h"




namespace quimeleon {





ItemRect::ItemRect(const Item* parent,
	const Glib::ustring& id, item_type_t type)
	: Item(parent, id, type),
	  _is_fixed(true),
	  _is_horizontal(true),
	  _spacing(0),
	  _is_transparent(true)
{
}




void
ItemRect::complete_attributes()
{
	if (_is_transparent) {
		set_alpha_normal(0);
	}
	align_children();

	Item::complete_attributes();
}



void
ItemRect::align_children()
{
	if (_is_fixed) {
		return;
	}

	const std::vector<std::vector<Item*> >& childs = children();
	if (childs.empty()) {
		return;
	}


	if (_is_horizontal) {
		int cnt = 0;
		int cnt_total = 0;
		float width_left = 100.0;
		for (std::vector<std::vector<Item*> >::const_iterator i = childs.begin();
		     i < childs.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {
				if ((*j)->visible()) {
					if ((*j)->width_percent()) {
						width_left -= (*j)->width_percent();
					} else {
						cnt++;
					}
					cnt_total++;
				}
			}
		}

		width_left -= _spacing * (cnt_total - 1);

		if (cnt > 0 && width_left > 0) {
			float d = width_left / cnt;
			float used_width = 0.0;
			for (std::vector<std::vector<Item*> >::const_iterator i
			     	= childs.begin();
			     i < childs.end(); i++) {
				for (std::vector<Item*>::const_iterator j = i->begin();
				     j < i->end(); j++) {
					if ((*j)->visible()) {
						(*j)->set_x(used_width, true);
						if ((*j)->width_percent()) {
							used_width += (*j)->width_percent();
						} else {
							(*j)->set_width(d, true, false);
							used_width += d;
						}
						used_width += _spacing;
					}
				}
			}
		} else {
			std::cerr << _("Cannot align children in box") << '\n';
		}
	} else {
		int cnt = 0;
		int cnt_total = 0;
		float height_left = 100.0;
		for (std::vector<std::vector<Item*> >::const_iterator i = childs.begin();
		     i < childs.end(); i++) {
			for (std::vector<Item*>::const_iterator j = i->begin();
			     j < i->end(); j++) {
				if ((*j)->visible()) {
					if ((*j)->height_percent()) {
						height_left -= (*j)->height_percent();
					} else {
						cnt++;
					}
					cnt_total++;
				}
			}
		}
		height_left -= _spacing * (cnt_total - 1);

		if (cnt > 0 && height_left > 0) {
			float d = height_left / cnt;
			float used_height = 0.0;
			for (std::vector<std::vector<Item*> >::const_iterator i
			     	= childs.begin();
			     i < childs.end(); i++) {
				for (std::vector<Item*>::const_iterator j = i->begin();
				     j < i->end(); j++) {
					const std::vector<ItemAnswer*>& child_answers
						= (*j)->answers();

					if (   int(child_answers.size()) >= alternative()
					    || j == i->end() - 1) {
						if ((*j)->visible()) {
							(*j)->set_y(used_height, true);
							if ((*j)->height_percent()) {
								used_height += (*j)->height_percent();
							} else {
								(*j)->set_height(d, true, false);
								used_height += d;
							}
							used_height += _spacing;
						}
						break;
					}
				}
			}
		} else {
			std::cerr << _("Cannot align children in box") << '\n';
		}
	}

}



void
ItemRect::compose_pixbuf()
{
	if (   ! visible()
	    || ! _pixbuf) {
		return;
	}

	if (   alpha() > 0
	    && crop_width() > 0
	    && crop_height() > 0) {
		Glib::RefPtr<Gdk::Drawable> drawable = Gdk::Pixmap::create(
			get_drawable(),
			width(), height(), -1);
		Glib::RefPtr<Gdk::Colormap> colormap = drawable->get_colormap();

		Gdk::Color fg_color;
		fg_color.set_rgb(r(), g(), b());
		fg_color.rgb_find_color(colormap);
		gc()->set_foreground(fg_color);

		drawable->draw_rectangle(gc(), true, 0, 0, crop_width(), crop_height());

		Glib::RefPtr<Gdk::Pixbuf> pixbuf
			= Gdk::Pixbuf::create(drawable, colormap,
				0, 0, 0, 0, crop_width(), crop_height());

		pixbuf->composite(_pixbuf,
			crop_x(), crop_y(), crop_width(), crop_height(),
			crop_x(), crop_y(), 1, 1,
			Gdk::INTERP_BILINEAR, alpha());
	}

	Item::compose_pixbuf();
}





} // end namespace quimeleon
