/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "item_status.h"

#include <limits>




namespace quimeleon {





ItemStatus::ItemStatus(Item* parent,
	const Glib::ustring& id, item_type_t type, int number,
	reference_t reference)
	: ItemMulti(parent, id, type, number),
	  _reference(reference),
	  _selected(std::numeric_limits<unsigned int>::max()),
	  _active(std::numeric_limits<unsigned int>::max()),
	  _is_next_level(false),
	  _is_next_question(false)
{
}




void
ItemStatus::complete_attributes()
{
	const std::vector<std::vector<Item*> >& childs = children();
	if (   _reference == R_next_level_and_question
	    && ! childs.empty()) {
		_active = 0;
		const std::vector<Item*>& childs_childs = childs[_active];
		for (std::vector<Item*>::const_iterator i = childs_childs.begin();
		     i < childs_childs.end(); i++) {
			(*i)->set_state_rec(active);
		}
	}

	ItemRect::complete_attributes();
}



void
ItemStatus::change_status()
{
	if (_selected != std::numeric_limits<unsigned int>::max()) {
		const std::vector<Item*>& childs_childs
			= _children_and_fillers[_selected];
		for (std::vector<Item*>::const_iterator i = childs_childs.begin();
		     i < childs_childs.end(); i++) {
			if (*i) {
				(*i)->set_state_rec(selected);
			}
		}
	}
	const std::vector<Item*>& childs_childs = _children_and_fillers[_active];
	for (std::vector<Item*>::const_iterator i = childs_childs.begin();
	     i < childs_childs.end(); i++) {
		if (*i) {
			(*i)->set_state_rec(active);
		}
	}

	if (restore_pixbuf()) {
		ItemRect::compose_pixbuf();

		signal_expose();
	}

}



void
ItemStatus::reset()
{
	const std::vector<std::vector<Item*> >& childs = children();
	for (std::vector<std::vector<Item*> >::const_iterator i = childs.begin();
	     i < childs.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_state_rec(normal);
		}
	}

	_active = _selected = std::numeric_limits<unsigned int>::max();
}



void
ItemStatus::set_level(unsigned int n)
{
	if (   _reference != R_level
	    || children().empty()) {
		return;
	}

	_selected = _active;
	_active = n % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::set_question(unsigned int n)
{
	if (   _reference != R_question
	    || children().empty()) {
		return;
	}

	_selected = _active;
	_active = n % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::next_level()
{
	if (   (   _reference != R_next_level
			&& _reference != R_next_level_and_question)
	    || children().empty()) {
		return;
	}

	if (_reference == R_next_level_and_question) {
		if (! _is_next_question) {
			_is_next_level = true;
			return;
		} else {
			_is_next_level = _is_next_question = false;
		}
	}
	_selected = _active;
	_active = (_active + 1) % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::next_question()
{
	if (   (_reference != R_next_question
			&& _reference != R_next_level_and_question)
	    || children().empty()) {
		return;
	}

	if (_reference == R_next_level_and_question) {
		if (! _is_next_level) {
			_is_next_question = true;
			return;
		} else {
			_is_next_level = _is_next_question = false;
		}
	}
	_selected = _active;
	_active = (_active + 1) % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::set_manual(unsigned int no, unsigned int n)
{
	if (   number() != no
	    || _reference != R_manual
	    || children().empty()) {
		return;
	}

	_selected = _active;
	_active = n % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::next_manual(unsigned int no)
{
	if (   number() != no
	    || _reference != R_manual
	    || children().empty()) {
		return;
	}

	_selected = _active;
	_active = (_active + 1) % _children_and_fillers.size();

	change_status();
}



void
ItemStatus::prev_manual(unsigned int no)
{
	if (   number() != no
	    || _reference != R_manual
	    || children().empty()) {
		return;
	}

	_selected = _active;
	_active = (_active - 1) % _children_and_fillers.size();

	change_status();
}



const std::vector<ItemStatus*>&
ItemStatus::statuss() const
{
	if (_statuss.empty()) {
		_statuss.push_back(const_cast<ItemStatus*>(this));
	}

	return _statuss;
}





} // end namespace quimeleon
