/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "item_top.h"

#include "item_answer.h"
#include "item_action.h"
#include "item_level.h"
#include "item_question.h"

#include "gettext.h"




namespace quimeleon {





ItemTop::ItemTop()
	: Item(0, "TOP", T_rect),
	  _mode(mode_black),
	  _level_cnt(0),
	  _parse_sequence(false),
	  _sequence_pos(0)
{
}



ItemTop::~ItemTop()
{
	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_intro.begin();
	     i < _children_intro.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_win.begin();
	     i < _children_outro_win.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_lose.begin();
	     i < _children_outro_lose.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
}




void
ItemTop::reset()
{
	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_intro.begin();
	     i < _children_intro.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
	_children_intro.clear();
	_pauses_intro.clear();

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_win.begin();
	     i < _children_outro_win.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
	_children_outro_win.clear();
	_pauses_outro_win.clear();

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_lose.begin();
	     i < _children_outro_lose.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			delete *j;
		}
	}
	_children_outro_lose.clear();
	_pauses_outro_lose.clear();


//	_mode = mode_black;
	_level_cnt = 0;
	_parse_sequence = false;
//	_sequence_pos = 0;

	Item::reset();
}



void
ItemTop::parse_sequence(bool parse_sequence, group_t group)
{

	switch (group) {
	case group_intro:
		_parse_sequence = parse_sequence;
		if (_parse_sequence) {
			_children_intro.push_back(std::vector<Item*>());
		} else if (_children_intro.back().empty()) {	// implies ! _children_intro.empty()
			_children_intro.pop_back();
		}
		break;
	case group_outro_win:
		_parse_sequence = parse_sequence;
		if (_parse_sequence) {
			_children_outro_win.push_back(std::vector<Item*>());
		} else if (_children_outro_win.back().empty()) {	// ...
			_children_outro_win.pop_back();
		}
		break;
	case group_outro_lose:
		_parse_sequence = parse_sequence;
		if (_parse_sequence) {
			_children_outro_lose.push_back(std::vector<Item*>());
		} else if (_children_outro_lose.back().empty()) {	// ...
			_children_outro_lose.pop_back();
		}
		break;
	default:
		throw ParserError(_("<sequence> not allowed in group trivia"));
	}
}



void
ItemTop::add_child(Item* child, group_t group)
{
	switch (group) {
	case group_intro:
		if (_parse_sequence) {
			_children_intro.back().push_back(child);
			_pauses_intro.push_back(0);
		} else {
			_children_intro.push_back(std::vector<Item*>(1, child));
		}
		break;
	case group_outro_win:
		if (_parse_sequence) {
			_children_outro_win.back().push_back(child);
			_pauses_outro_win.push_back(0);
		} else {
			_children_outro_win.push_back(std::vector<Item*>(1, child));
		}
		break;
	case group_outro_lose:
		if (_parse_sequence) {
			_children_outro_lose.back().push_back(child);
			_pauses_outro_lose.push_back(0);
		} else {
			_children_outro_lose.push_back(std::vector<Item*>(1, child));
		}
		break;
	default:
		Item::add_child(child);
	}
}



void
ItemTop::add_pause(int seconds, group_t group)
{
	switch (group) {
	case group_intro:
		if (_pauses_intro.empty()) {
			throw ParserStyleError(_("use <pause> after <item>"));
		}
		_pauses_intro.back() = seconds;
		break;
	case group_outro_win:
		if (_pauses_outro_win.empty()) {
			throw ParserStyleError(_("use <pause> after <item>"));
		}
		_pauses_outro_win.back() = seconds;
		break;
	case group_outro_lose:
		if (_pauses_outro_win.empty()) {
			throw ParserStyleError(_("use <pause> after <item>"));
		}
		_pauses_outro_lose.back() = seconds;
		break;
	default:
		throw ParserError(_("<sequence> not allowed in group trivia"));
	}
}



void
ItemTop::draw(int x, int y, int w, int h, bool do_expose)
{
	if (_pixbuf) {

		if (   w == -1
		    || w + x > _pixbuf->get_width()) {
			w = _pixbuf->get_width() - x;
		}
		if (   h == -1
		    || h + y > _pixbuf->get_height()) {
			h = _pixbuf->get_height() - y;
		}

		if (do_expose) {
			compose_pixbuf();
		}
		_drawable->draw_pixbuf(gc(), _pixbuf, x, y,
			x, y, w, h,
			Gdk::RGB_DITHER_NORMAL, x, y);
	}
}



void
ItemTop::set_parent_dimensions(int parent_x, int parent_y,
	int parent_w, int parent_h)
{
	set_font_scale(parent_w, parent_h);

	Item::set_parent_dimensions(parent_x, parent_y, parent_w, parent_h,
		parent_x, parent_y, parent_w, parent_h);

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_intro.begin();
	     i < _children_intro.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_parent_dimensions(x(), y(), width(), height(),
				x(), y(), width(), height());
		}
	}

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_win.begin();
	     i < _children_outro_win.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_parent_dimensions(x(), y(), width(), height(),
				x(), y(), width(), height());
		}
	}

	for (std::vector<std::vector<Item*> >::const_iterator i
	     	= _children_outro_lose.begin();
	     i < _children_outro_lose.end(); i++) {
		for (std::vector<Item*>::const_iterator j = i->begin();
		     j < i->end(); j++) {
			(*j)->set_parent_dimensions(x(), y(), width(), height(),
				x(), y(), width(), height());
		}
	}


	_pixbuf = Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, false, 8,
		parent_w, parent_h);

	//compose_pixbuf();
}



void
copy_action_state(const std::vector<ItemAction*>& a,
	const std::vector<ItemAction*>& b)
{
	std::vector<ItemAction*>::size_type i;
	std::vector<ItemAction*>::size_type j;

	for (i = j = 0; i < a.size() && j < b.size();) {
		 ItemAction::cmp_id action_cmp;
		int cmp = action_cmp(a[i], b[j]);
		if (cmp == 0) {
			b[j]->mark_state(a[i]->state(), false);
			i++;
			j++;
		} else if (cmp > 0) {
			i++;
		} else {
			j++;
		}
	}
}



void
copy_level_state(const std::vector<ItemLevel*>& a,
	const std::vector<ItemLevel*>& b)
{
	std::vector<ItemLevel*>::size_type i;
	std::vector<ItemLevel*>::size_type j;

	for (i = j = 0; i < a.size() && j < b.size();) {
		 ItemLevel::cmp_id level_cmp;
		int cmp = level_cmp(a[i], b[j]);
		if (cmp == 0) {
			b[j]->mark_state(a[i]->state(), false);
			i++;
			j++;
		} else if (cmp > 0) {
			i++;
		} else {
			j++;
		}
	}
}



unsigned int
ItemTop::set_question(const Glib::ustring& question_str,
	const std::vector<answer_t>& answer_datas,
	unsigned int correct, bool& alternative_changed)
{
	unsigned int correct_pos = correct;
	bool do_compose = true;

	answer_type_t preferred_type = AT_both;
	for (std::vector<answer_t>::const_iterator i = answer_datas.begin();
	     i < answer_datas.end(); i++) {
		if (! i->text.empty()) {
			if (preferred_type == AT_image) {
				preferred_type = AT_both;
				break;
			} else {
				preferred_type = AT_text;
			}
		}
		if (! i->image.empty()) {
			if (preferred_type == AT_text) {
				preferred_type = AT_both;
				break;
			} else {
				preferred_type = AT_image;
			}
		}
	}


	const std::vector<ItemAction*> actions_old = actions();
	const std::vector<ItemLevel*> levels_old = levels();
	alternative_changed = set_alternative(answer_datas.size(), preferred_type);
	const std::vector<ItemAnswer*>& item_answers = answers();

	if (alternative_changed) {
		do_compose = false;
		const std::vector<ItemAction*> actions_new = actions();
		if (actions_old != actions_new) {
			copy_action_state(actions_old, actions_new);
		}
		const std::vector<ItemLevel*> levels_new = levels();
		if (levels_old != levels_new) {
			copy_level_state(levels_old, levels_new);
		}

		std::vector<ItemAnswer*>::size_type last_used
			= (answer_datas.size() < item_answers.size()) ?
				answer_datas.size() : item_answers.size();
		for (std::vector<ItemAnswer*>::size_type i = 0;
		     i < last_used; i++) {
			item_answers[i]->set_is_used(true);
		}
		for (std::vector<ItemAnswer*>::size_type i = last_used;
		     i < item_answers.size(); i++) {
			item_answers[i]->set_is_used(false);
		}
	}

	std::vector<ItemAnswer*>::size_type j = 0;
	if (   item_answers.empty()
	    && ! answer_datas.empty()) {
		char* type_str;
		switch (preferred_type) {
		case AT_text:
			type_str = _("text only");
			break;
		case AT_image:
			type_str = _("images only");
			break;
		default:
			type_str = _("text and images");
		}
		char errstr[127];
		snprintf(errstr, 127, _("No items for answers (%s)"), type_str);

		throw LayoutError(errstr);
	}
	if (   item_answers.size() < answer_datas.size()
	    && correct >= item_answers.size()) {
		correct_pos = correct % item_answers.size();
		for (std::vector<answer_t>::const_iterator i = answer_datas.begin();
		     i < answer_datas.end() && j < item_answers.size(); i++, j++) {
			if (j == correct_pos) {
				item_answers[j]->set_text_and_image(
					answer_datas[correct].text,
					answer_datas[correct].image,
					do_compose);
				item_answers[j]->mark_state(normal, do_compose);
			} else {
				item_answers[j]->set_text_and_image(i->text, i->image,
				                                    do_compose);
				item_answers[j]->mark_state(normal, do_compose);
			}
		}
	} else {
		for (std::vector<answer_t>::const_iterator i = answer_datas.begin();
		     i < answer_datas.end() && j < item_answers.size(); i++, j++) {
			item_answers[j]->set_text_and_image(i->text, i->image, do_compose);
			item_answers[j]->mark_state(normal, do_compose);
		}
	}

	ItemQuestion* item_question = question();
	if (item_question) {
		item_question->set_text(question_str, do_compose);
	} else {
		std::cerr << "Nanananu?\n";
	}

	if (! do_compose) {
		compose_pixbuf();
		signal_expose();
	}


	return correct_pos;
}



unsigned int
ItemTop::set_level_cnt(unsigned int level_cnt)
{
	const std::vector<ItemLevel*>& the_levels = all_levels();
	if (level_cnt > levels().size()) {
		_level_cnt = levels().size();
	} else {
		_level_cnt = level_cnt;
	}
	if (level_cnt == 0) {
		throw LayoutError(_("no levels"));
	}


// use same formula as in mark_state_level()
	float pos_fac = float(levels().size()) / _level_cnt;
	std::vector<Item*> parents;

	for (unsigned int i = 0, j = 0, l = 0; i < _level_cnt; i++, j++) {
		unsigned int k = int((0.9 + i) * pos_fac);
		while (k > j) {
			do {
				if (the_levels[l]->visible()) {
					the_levels[l]->set_visible(false);
					if (   parents.empty()
						|| the_levels[l]->parent()
							!= parents[parents.size() - 1]) {
						parents.push_back(the_levels[l]->parent());
					}
				}
			} while (   ++l < the_levels.size()
					 && ItemLevel::cmp_id()(
								the_levels[l - 1], the_levels[l]) == 0);

			if (l >= the_levels.size()) {
				break;
			}
			j++;
		}
		do {
			if (! the_levels[l]->visible()) {
				the_levels[l]->set_visible(true);
				if (   parents.empty()
					|| the_levels[l]->parent()
						!= parents[parents.size() - 1]) {
					parents.push_back(the_levels[l]->parent());
				}
			}
			the_levels[l]->set_label_number(i + 1);
		} while (   ++l < the_levels.size()
				 && ItemLevel::cmp_id()(
						 the_levels[l - 1], the_levels[l]) == 0);
	}

	for (std::vector<Item*>::iterator i = parents.begin(); i < parents.end();
		 i++) {
		if (ItemRect* item_rect = dynamic_cast<ItemRect*>(*i)) {
			item_rect->align_children();
		}
	}

	return _level_cnt;
}



int
ItemTop::set_mode(mode_t mode)
{
	_mode = mode;

	switch (_mode) {
	case mode_intro:
	case mode_outro_win:
	case mode_outro_lose:
		return set_sequence_pos(0);
	default:
		compose_pixbuf();
		signal_expose();
	}

	return -1;
}



int
ItemTop::set_sequence_pos(unsigned int n)
{
	switch (_mode) {
	case mode_intro:
		if (   ! _children_intro.empty()
		    && n < _children_intro[0].size()) {
			_sequence_pos = n;
			compose_pixbuf();
			signal_expose();

			if (_pauses_intro.size() > _sequence_pos) {
				return _pauses_intro[_sequence_pos];
			} else {
				return 0;
			}
		}

		return -1;
	case mode_outro_win:
		if (   ! _children_outro_win.empty()
		    && n < _children_outro_win[0].size()) {
			_sequence_pos = n;

			compose_pixbuf();
			signal_expose();

			if (_pauses_outro_win.size() > _sequence_pos) {
				return _pauses_outro_win[_sequence_pos];
			} else {
				return 0;
			}
		}

		return -1;
	case mode_outro_lose:
		if (   ! _children_outro_lose.empty()
		    && n < _children_outro_lose[0].size()) {
			_sequence_pos = n;

			compose_pixbuf();
			signal_expose();

			if (_pauses_outro_lose.size() > _sequence_pos) {
				return _pauses_outro_lose[_sequence_pos];
			} else {
				return 0;
			}
		}

		return -1;
	default:
		return -1;
	}
}



void
ItemTop::mark_state_level(unsigned int level, state_t state)
{
	const std::vector<ItemLevel*>& the_levels = levels();

// use same formula as in set_level_cnt()
	float pos_fac = float(the_levels.size()) / float(_level_cnt);
	int k = int((float(level) + 0.9) * pos_fac);

	the_levels[k]->mark_state(state);
}



ItemArea*
ItemTop::get_area(const Glib::ustring& id)
{
	std::map<const Glib::ustring, ItemArea*> child_areas = areas();
	std::map<const Glib::ustring, ItemArea*>::iterator i = child_areas.find(id);

	if (i == child_areas.end()) {
		return 0;
	} else {
		return i->second;
	}
}



void
ItemTop::compose_pixbuf()
{
	if (! visible() || ! _pixbuf || ! gc()) {
		return;
	}

	if (alpha() > 0) {
		Glib::RefPtr<Gdk::Drawable> drawable = Gdk::Pixmap::create(
			get_drawable(), width(), height(), -1);
		Glib::RefPtr<Gdk::Colormap> colormap = drawable->get_colormap();

		Gdk::Color white("white");
		Gdk::Color black("black");
		white.rgb_find_color(colormap);
		black.rgb_find_color(colormap);
		gc()->set_foreground(black);
		gc()->set_background(white);
		drawable->draw_rectangle(gc(), true, 0, 0, width(), height());
		Glib::RefPtr<Gdk::Pixbuf> pixbuf
			= Gdk::Pixbuf::create(drawable, colormap,
				0, 0, 0, 0, width(), height());

		pixbuf->composite(_pixbuf, x(), y(), width(), height(), x(), y(), 1, 1,
			Gdk::INTERP_BILINEAR, 0xff);
	}


	switch (_mode) {
	case mode_black:
		break;
	case mode_intro:
		for (std::vector<std::vector<Item*> >::const_iterator i
		     	 = _children_intro.begin();
		     i < _children_intro.end(); i++) {
			if (_sequence_pos < i->size()) {
				(*i)[_sequence_pos]->compose_pixbuf();
			} else {
				i->back()->compose_pixbuf();
			}
		}
		break;
	case mode_outro_win:
		for (std::vector<std::vector<Item*> >::const_iterator i
		     	 = _children_outro_win.begin();
		     i < _children_outro_win.end(); i++) {
			if (_sequence_pos < i->size()) {
				(*i)[_sequence_pos]->compose_pixbuf();
			} else {
				i->back()->compose_pixbuf();
			}
		}
		break;
	case mode_outro_lose:
		for (std::vector<std::vector<Item*> >::const_iterator i
		     	 = _children_outro_lose.begin();
		     i < _children_outro_lose.end(); i++) {
			if (_sequence_pos < i->size()) {
				(*i)[_sequence_pos]->compose_pixbuf();
			} else {
				i->back()->compose_pixbuf();
			}
		}
		break;
	default:
		Item::compose_pixbuf();
	}
}





} // end namespace quimeleon
