/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "../config.h"

#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <gtkmm/main.h>
#include <gtkmm/messagedialog.h>
#include <glibmm/thread.h>

#include "error.h"
#include "gettext.h"
#include "parser.h"
#include "preferences.h"
#include "quimeleon.h"
#include "trivias.h"




void
usage(const char* program)
{
	char opt_descr[255];
	snprintf(opt_descr, 255, _("Search for bitmaps, sounds, style-file, and\n                             trivia-file in %sNAME\n                             before global data directory"),
		DATADIRROOT);

	std::cout << _("Usage: ") << program << " [" << _("options") << "] [" << _("DATADIR") << " ...]\n"
	          << _("  when given, search for bitmaps, sounds, style-file, and trivia-file\n  in DATADIRs after your current working directory and before all other\n  directories.") << '\n'
	          << _("  For more than one DATADIR, every additional DATADIR takes precedence\n  over other directories.") << '\n'
	          << "  -h, --help                 "
	          << _("Show this help.") << '\n'
	          << "  -v, --version              "
	          << _("Show version of program.") << '\n'
	          << "  -t" << _("FILE") << ", --triviafile=" << _("FILE") << "  "
	          << _("Read questions and answers from FILE.") << '\n'
	          << "  -s" << _("FILE") << ", --stylefile=" << _("FILE") << "   "
	          << _("Read style data from FILE.") << '\n'
	          << "  -S" << _("NAME") << ", --style=" << _("NAME") << "       "
	          << opt_descr << '\n'
	          << "  -f, --fullscreen           "
	          << _("Start in fullscreen mode.") << '\n'
			  << "  -n, --no-buffer            "
	          << _("Do not keep Pixmaps in RAM.") << '\n'
			  << "  -b, --buffer               "
	          << _("Keep Pixmaps in RAM.") << ' ' << _("(default)")
			  << "\n\n";
}




int
main(int argc, char* argv[])
{
	try {
		std::locale::global(std::locale(""));
	} catch (const std::exception& e) {
		std::cerr << argv[0] << ": " << e.what() << '\n';
	}
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);


	Gtk::Main app(argc, argv);
	try {
		quimeleon::Configuration configuration;
		configuration.read_options();

		int opt_ch;
		struct option longopts[9] = {
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'v' },
			{ "triviafile", 1, 0, 't' },
			{ "stylefile", 1, 0, 's' },
			{ "style", 1, 0, 'S' },
			{ "fullscreen", 2, 0, 'f' },
			{ "buffer", 2, 0, 'b' },
			{ "no-buffer", 0, 0, 'n' },
			{ 0, 0, 0, 0 }
		};
		while ((opt_ch = getopt_long(argc, argv, "hvt:s:S:fbn", longopts, 0))
		       	!= -1)
		{
			switch (opt_ch) {
			case '?':
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'v':
				std::cout << PACKAGE << " v" << VERSION << '\n'
				          << _("Written by Aike Reyer") << "\n\n"
				          << _("Copyright (C) 2003, 2005, 2006 Aike Reyer <aike@users.sourceforge.net>") << '\n'
				          << _("This is free software; see the source for copying conditions.  There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.")
				          << '\n';
				exit(0);
			case 't':
				configuration.set_trivia_file(optarg);
				break;
			case 's':
				configuration.set_style_file(optarg);
				break;
			case 'S':
				configuration.set_style_name(optarg);
				break;
			case 'f':
				configuration.set_fullscreen(
					! optarg || std::string(optarg) == "true");
				break;
			case 'n':
				configuration.set_do_buffer(false);
				break;
			case 'b':
				configuration.set_do_buffer(
					! optarg || std::string(optarg) == "true");
				break;
			default:
				std::cerr << char(opt_ch) << ": ?\n";
			}
		}

		for (int i = optind; i < argc; i++) {
			configuration.add_data_dir(argv[i]);
		}


		bool is_done = false;
		quimeleon::Trivias trivias;
		do {
			try {
				// use temporary Configuration to avoid multiple initialization
				// of style
				quimeleon::Configuration tmp_configuration = configuration;
				quimeleon::parse_style(tmp_configuration);

				trivias = quimeleon::Trivia::read_trivia_file(
					configuration.trivia_file().c_str(), configuration);

				is_done = true;
			} catch (const quimeleon::Exception& e) {
				Gtk::MessageDialog message_dialog(
					Glib::locale_to_utf8(e.what()), false, Gtk::MESSAGE_ERROR);
				message_dialog.run();
				message_dialog.hide();

				quimeleon::Preferences preferences(&configuration, false);
				int re = preferences.run();

				if (re == Gtk::RESPONSE_CANCEL) {
					return 0;
				}
			}

		} while (! is_done);

#ifdef DEFAULT_ICON
		try {
			Gtk::Window::set_default_icon_from_file(DEFAULT_ICON);
		} catch (...) {
		}
#endif

		quimeleon::parse_style(configuration);
		quimeleon::Quimeleon quimeleon(&configuration, trivias);

		quimeleon.show();

		app.run(quimeleon);

	} catch (int e) {
		std::cerr << argv[0] << ": " << _("unexpected error")
			<< ' ' << e << '\n';
	} catch (const quimeleon::Exception& e) {
		std::cerr << argv[0] << ": " << e.what() << '\n';
		Gtk::MessageDialog message_dialog(
			Glib::locale_to_utf8(e.what()), false, Gtk::MESSAGE_ERROR);
		message_dialog.run();
	} catch (const Glib::Exception& e) {
		std::cerr << argv[0] << "[Glib]: " << e.what() << '\n';
	} catch (const std::exception& e) {
		std::cerr << argv[0] << ": " << e.what() << '\n';
	} catch (...) {
		std::cerr << argv[0] << ": " << _("unexpected error") << '\n';
	}


	return 0;
}

