/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "preferences.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

#include <gtkmm/filechooserdialog.h>
#include <gtkmm/frame.h>
#include <gtkmm/label.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/separator.h>
#include <gtkmm/stock.h>
#include <gtkmm/table.h>

#include "gettext.h"




namespace quimeleon {





Preferences::Preferences(Gtk::Window& parent, Configuration* configuration)
	: Gtk::Dialog(_8("Preferences"), parent),
	  _configuration(configuration),
	  _is_new_style(false),
	  _is_new_trivia(false)
{
	init();
}



Preferences::Preferences(Configuration* configuration, bool do_save)
	: Gtk::Dialog(_8("Preferences")),
	  _configuration(configuration),
	  _is_new_style(false),
	  _is_new_trivia(false)
{
	init(do_save);
}



void
Preferences::init(bool do_save)
{
	_old_style_name = _configuration->style_name();
	try {
		_old_trivia_file = _configuration->trivia_file();
	} catch (...) {
	}
	try {
		_old_style_file = _configuration->style_file();
	} catch (...) {
	}



	Gtk::VBox* vbox = get_vbox();

	Gtk::Table* table = Gtk::manage(new Gtk::Table(4, 3));

	Gtk::Label* label;


	/// available styles in DATADIRROOT

	try {
		Glib::Dir directory(DATADIRROOT);

		label = Gtk::manage(new Gtk::Label(_8("_Style:"),
		                    Gtk::ALIGN_LEFT, Gtk::ALIGN_CENTER, true));
		table->attach(*label, 0, 1, 0, 1,
			Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK, 1, 2);

		_combo_box_style = Gtk::manage(new Gtk::ComboBoxText);

		_combo_box_style->append_text("");

		struct stat file_stat;

		for (Glib::Dir::const_iterator i = directory.begin();
		     i != directory.end(); i++) {
			if (stat(Glib::build_filename(DATADIRROOT, *i).c_str(),
			         &file_stat) != 0) {
				continue;
			}
			if (   S_ISDIR(file_stat.st_mode)
			    && (*i)[0] != '.') {
				_combo_box_style->append_text(*i);
			}
		}


		_combo_box_style->set_active_text(_configuration->style_name());

		label->set_mnemonic_widget(*_combo_box_style);
		table->attach(*_combo_box_style, 1, 2, 0, 1,
			Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK, 1, 2);
	} catch (const Glib::Error& e) {
		_combo_box_style = false;
		std::cerr << e.what() << '\n';
	}


	/// style file

	label = Gtk::manage(new Gtk::Label(_8("St_yle file:"),
			            Gtk::ALIGN_LEFT, Gtk::ALIGN_CENTER, true));
	table->attach(*label, 0, 1, 1, 2,
		Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK, 1, 2);
	_entry_style_file = Gtk::manage(new Gtk::Entry);
	try {
		_entry_style_file->set_text(
			Glib::filename_to_utf8(_configuration->style_file_name()));
	} catch (const Glib::Error& e) {
		_entry_style_file->set_text(_configuration->style_file_name());
	}
	table->attach(*_entry_style_file, 1, 2, 1, 2,
		Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK, 1, 2);
	Gtk::Button* button = Gtk::manage(new Gtk::Button(_8("Browse ...")));
	button->signal_clicked().connect(
		sigc::mem_fun(*this, &Preferences::style_file_select));
	label->set_mnemonic_widget(*button);
	table->attach(*button, 2, 3, 1, 2,
		Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK, 1, 0);


	/// trivia file

	label = Gtk::manage(new Gtk::Label(_8("_Trivia file:"),
			            Gtk::ALIGN_LEFT, Gtk::ALIGN_CENTER, true));
	table->attach(*label, 0, 1, 2, 3,
		Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK, 1, 2);
	_entry_trivia_file = Gtk::manage(new Gtk::Entry);
	try {
		_entry_trivia_file->set_text(
			Glib::filename_to_utf8(_configuration->trivia_file_name()));
	} catch (const Glib::Error& e) {
		_entry_trivia_file->set_text(_configuration->trivia_file_name());
	}
	table->attach(*_entry_trivia_file, 1, 2, 2, 3,
		Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK, 1, 2);
	button = Gtk::manage(new Gtk::Button(_8("Browse ...")));
	button->signal_clicked().connect(
		sigc::mem_fun(*this, &Preferences::trivia_file_select));
	label->set_mnemonic_widget(*button);
	table->attach(*button, 2, 3, 2, 3,
		Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK, 1, 0);

	table->show_all();

	vbox->pack_start(*table);


	Gtk::HSeparator* separator = Gtk::manage(new Gtk::HSeparator);
	separator->show();
	vbox->pack_start(*separator, Gtk::PACK_SHRINK, 4);


	/// list of directories with style data

	Gtk::Frame* frame = Gtk::manage(new Gtk::Frame(_8("Data dirs")));

	_data_dirs_editor = Gtk::manage(new DataDirsEditor(_configuration));

	label = dynamic_cast<Gtk::Label*>(frame->get_label_widget());
	if (label) {
		label->set_text_with_mnemonic(_8("_Data dirs"));
		label->set_mnemonic_widget(_data_dirs_editor->list_view());
	}


	Gtk::HBox* hbox = Gtk::manage(new Gtk::HBox);
	hbox->set_border_width(2);

	Gtk::ScrolledWindow* scrolled_window = Gtk::manage(new Gtk::ScrolledWindow);
	scrolled_window->set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC);

	scrolled_window->add(_data_dirs_editor->list_view());

	hbox->pack_start(*scrolled_window);
	hbox->pack_start(_data_dirs_editor->button_box(), Gtk::PACK_SHRINK);

	frame->add(*hbox);

	frame->show_all();

	table->attach(*frame, 0, 3, 3, 4,
		Gtk::EXPAND | Gtk::FILL, Gtk::EXPAND | Gtk::FILL, 1, 2);


	/// fullscreen

	_check_button_fullscreen
		= Gtk::manage(new Gtk::CheckButton(_8("start in _fullscreen mode"),
		                                   true));
	if (_configuration->fullscreen()) {
		_check_button_fullscreen->set_active();
	}
	_check_button_fullscreen->show();
	vbox->pack_start(*_check_button_fullscreen, Gtk::PACK_SHRINK);


	/// do buffer

	_check_button_do_buffer
		= Gtk::manage(new Gtk::CheckButton(_8("buffer _graphics"), true));
	if (_configuration->do_buffer()) {
		_check_button_do_buffer->set_active();
	}
	_check_button_do_buffer->show();
	vbox->pack_start(*_check_button_do_buffer, Gtk::PACK_SHRINK);


	/// do show hint

	_check_button_do_show_hint
		= Gtk::manage(new Gtk::CheckButton(_8("show _hint at startup"), true));
	if (_configuration->do_show_hint()) {
		_check_button_do_show_hint->set_active();
	}
	_check_button_do_show_hint->show();
	vbox->pack_start(*_check_button_do_show_hint, Gtk::PACK_SHRINK);


	add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	add_button(do_save ? Gtk::Stock::SAVE : Gtk::Stock::OK, Gtk::RESPONSE_OK);
}




void
Preferences::on_response(int response_id)
{
	switch (response_id) {
	case Gtk::RESPONSE_CANCEL:
		break;
	case Gtk::RESPONSE_OK:
		std::vector<std::string> data_dirs = _data_dirs_editor->data_dirs();

		if (data_dirs.size() != _configuration->data_dirs().size()) {
			_is_new_style = true;
		} else {
			const std::vector<std::string>& old_data_dirs
				= _configuration->data_dirs();
			for (std::vector<std::string>::size_type i = 0;
			     i < data_dirs.size(); i++) {
				if (data_dirs[i] != old_data_dirs[i]) {
					_is_new_style = true;
					break;
				}
			}
		}


		_configuration->data_dirs_clear();

		std::vector<std::string>::size_type global_data_dirs_size
			= _configuration->global_data_dirs_size();

		for (std::vector<std::string>::size_type i = global_data_dirs_size;
		     i < data_dirs.size(); i++) {
			_configuration->add_data_dir(data_dirs[i]);
		}

		Glib::ustring candidate;
		if (_combo_box_style) {
			try {
				candidate = Glib::filename_from_utf8(
						_combo_box_style->get_active_text());
			} catch (const Glib::Error& e) {
				candidate = _combo_box_style->get_active_text();
			} 
			if (candidate != _old_style_name) {
				_configuration->set_style_name(candidate);
				_is_new_style = true;
			}
		}

		try {
			candidate = Glib::filename_from_utf8(_entry_style_file->get_text());
		} catch (const Glib::Error& e) {
			candidate = _entry_style_file->get_text();
		}
		try {
			if (find_file(candidate) != _old_style_file) {
				_configuration->set_style_file(candidate);
				_is_new_style = true;
			}
		} catch (const FileError& e) {
			_configuration->set_style_file(candidate);
			_is_new_style = true;
		}

		try {
			candidate =
				Glib::filename_from_utf8(_entry_trivia_file->get_text());
		} catch (const Glib::Error& e) {
			candidate = _entry_trivia_file->get_text();
		}
		try {
			if (find_file(candidate) != _old_trivia_file) {
				_configuration->set_trivia_file(candidate);
				_is_new_trivia = true;
			}
		} catch (const FileError& e) {
			_configuration->set_trivia_file(candidate);
			_is_new_trivia = true;
		}


		bool do_buffer = _check_button_do_buffer->get_active();
		if (do_buffer != _configuration->do_buffer()) {
			_configuration->set_do_buffer(do_buffer);
		}

		bool do_show_hint = _check_button_do_show_hint->get_active();
		if (do_show_hint != _configuration->do_show_hint()) {
			_configuration->set_do_show_hint(do_show_hint);
		}


		_configuration->set_fullscreen(_check_button_fullscreen->get_active());


		break;
	}
}



void
Preferences::style_file_select()
{
	Gtk::FileChooserDialog file_chooser_dialog(_8("Select style file"));
	file_chooser_dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	file_chooser_dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	Gtk::FileFilter file_filter;
	file_filter.set_name(_8("Quimeleon style"));
	file_filter.add_pattern("*.style");
	file_chooser_dialog.add_filter(file_filter);


	Glib::ustring old_file;
	try {
		old_file = find_file(_entry_style_file->get_text());
		file_chooser_dialog.set_filename(absolute_path(old_file));
	} catch (const FileNotFound& e) {
		try {
			file_chooser_dialog.set_filename(
				absolute_path(_configuration->style_file()));
		} catch (const FileNotFound& e) {
		}
	}

	if (file_chooser_dialog.run() == Gtk::RESPONSE_OK) {
		std::string new_file = file_chooser_dialog.get_filename();

		if (new_file != old_file) {

			// truncate path if possible
			std::string basename = Glib::path_get_basename(new_file);
			if (basename != new_file) {
				try {
					std::string other_file = find_file(basename);

					if (other_file == new_file) {
						new_file = basename;
					}
				} catch (const FileNotFound& e) {
				}
			}
			_entry_style_file->set_text(new_file);
		}
	}
}



void
Preferences::trivia_file_select()
{
	Gtk::FileChooserDialog file_chooser_dialog(
		_8("Select trivia file"));
	file_chooser_dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	file_chooser_dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	Gtk::FileFilter file_filter;
	file_filter.set_name(_8("Quimeleon trivia"));
	file_filter.add_pattern("*.trivia");
	file_chooser_dialog.add_filter(file_filter);


	std::string old_file;
	try {
		old_file = find_file(_entry_trivia_file->get_text());
		file_chooser_dialog.set_filename(absolute_path(old_file));
	} catch (const FileNotFound& e) {
		try {
			file_chooser_dialog.set_filename(
				absolute_path(_configuration->trivia_file()));
		} catch (const FileNotFound& e) {
		}
	}

	if (file_chooser_dialog.run() == Gtk::RESPONSE_OK) {
		std::string new_file = file_chooser_dialog.get_filename();

		if (new_file != old_file) {

			// truncate path if possible
			std::string basename = Glib::path_get_basename(new_file);
			if (basename != new_file) {
				try {
					std::string other_file = find_file(basename);

					if (other_file == new_file) {
						new_file = basename;
					}
				} catch (const FileNotFound& e) {
				}
			}
			_entry_trivia_file->set_text(new_file);
		}
	}
}



std::string
Preferences::find_file(const std::string& filename) const
{
    struct stat info;

    if (stat(filename.c_str(), &info) != -1) {
        return filename;
    }


	if (_combo_box_style) {
		std::string style_name = _combo_box_style->get_active_text();

		if (! style_name.empty()) {
			std::vector<std::string> path(4);
			path[0] = DATADIRROOT;
			path[1] = style_name;
			path[2] = _configuration->lang();
			path[3] = filename;
	        std::string re = Glib::build_filename(path);
	        if (stat(re.c_str(), &info) != -1) {
	            return re;
	        }
			path.resize(3);
			path[2] = filename;
	        re = Glib::build_filename(path);
	        if (stat(re.c_str(), &info) != -1) {
	            return re;
	        }
	    }
	}


	std::vector<std::string> path3(3);
	std::vector<std::string> path2(2);
	path3[1] = _configuration->lang();
	path3[2] = path2[1] = filename;

	std::vector<std::string> data_dirs = _data_dirs_editor->data_dirs();
    for (std::vector<std::string>::reverse_iterator i
            = data_dirs.rbegin();
         i < data_dirs.rend(); i++) {
		path3[0] = *i;
        std::string re = Glib::build_filename(path3);
        if (stat(re.c_str(), &info) != -1) {
            return re;
        }
		path2[0] = *i;
        re = Glib::build_filename(path2);
        if (stat(re.c_str(), &info) != -1) {
            return re;
        }
    }


    throw FileNotFound(filename);
}





} // end namespace quimeleon
