/*
 * Quimeleon - a themeable quiz engine
 * Copyright (C) 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "quimeleon.h"

#include <gtkmm/main.h>
#include <gtkmm/messagedialog.h>
#include <gdkmm/cursor.h>
#include <gdk/gdkkeysyms.h>

#include "../config.h"

#include "item_action.h"
#include "item_level.h"
#include "item_answer.h"
#include "welcome.h"

#include "gettext.h"




namespace quimeleon {






static Gdk::Cursor
hidden_mouse_cursor()
{
	const char bm_null = 0x00;
	Gdk::Color gblack("black");
	Glib::RefPtr<Gdk::Bitmap> bm_off = Gdk::Bitmap::create(&bm_null, 1, 1);

	return Gdk::Cursor(bm_off, bm_off, gblack, gblack, 0, 0);
}



Quimeleon::Quimeleon(Configuration* configuration, const Trivias& trivias)
	: _configuration(configuration),
	  _control(configuration, trivias),
	  _plugins(configuration, &_control),
	  _is_fullscreen(configuration->fullscreen()),
	  _display(_configuration->item_top()),
	  _ui_manager(UIManager::create()),
	  _dialogs(this, configuration, _display, _control, _plugins)
{
	set_title(PACKAGE_NAME);
	set_size_request(100, 80);

	_menu = _ui_manager->get_widget("/MainMenu");
	_toolbar = _ui_manager->get_widget("/ToolBar");
	set_default_size(480,
		360 + _menu->size_request().height + _toolbar->size_request().height);

	_vbox.pack_start(*_menu, Gtk::PACK_SHRINK);
	_vbox.pack_start(*_toolbar, Gtk::PACK_SHRINK);

	add_accel_group(_ui_manager->get_accel_group());
	_ui_manager->signal_file_open().connect(
		sigc::mem_fun(_dialogs, &Dialogs::file_open));
	_ui_manager->signal_file_save().connect(
		sigc::mem_fun(_dialogs, &Dialogs::save_style_and_trivia));
	_ui_manager->signal_file_quit().connect(
		sigc::mem_fun(*this, &Gtk::Window::hide));
	_ui_manager->signal_settings_preferences().connect(
		sigc::mem_fun(_dialogs, &Dialogs::preferences));
	_ui_manager->signal_help_about().connect(
		sigc::mem_fun(_dialogs, &Dialogs::about));

	_vbox.pack_end(_display);
	add(_vbox);

	show_all();

	_display.add_events(  Gdk::ENTER_NOTIFY_MASK
	                    | Gdk::LEAVE_NOTIFY_MASK
	                    | Gdk::POINTER_MOTION_MASK);
	_display.signal_enter_notify_event().connect(
		sigc::mem_fun(this, &Quimeleon::start_mouse_watch));
	_display.signal_leave_notify_event().connect(
		sigc::mem_fun(this, &Quimeleon::stop_mouse_watch));


	if (_is_fullscreen) {
		fullscreen();
	}

}




bool
Quimeleon::on_key_press_event(GdkEventKey* key)
{
	if (   key->state & (  GDK_CONTROL_MASK
	                     | GDK_MOD1_MASK	// Alt
	                     | GDK_MOD4_MASK)	// Super
		&& Window::on_key_press_event(key)) {
		return true;
	}

	try {
	command_t command = _configuration->keys().command(*key);

	if (   command >= command_select_answer_begin
	    && command <= command_select_answer_end) {
		const unsigned int n = command - command_select_answer_begin;
		if (_control.answer_selected(n)) {
			_control.set_gamestate(STATE_answer);
		}
	} else if (   command >= command_deselect_answer_begin
	           && command <= command_deselect_answer_end) {
		const unsigned int n = command - command_deselect_answer_begin;
		_control.answer_normal(n);
	} else if (   _control.gamestate() >= STATE_question
	           && _control.gamestate() <= STATE_show_answer
	           && command >= command_select_action_begin
	           && command <= command_select_action_end) {
		const std::vector<ItemAction*>& actions =
			_configuration->item_top()->actions();
		const std::vector<ItemAction*>::size_type n
			= command - command_select_action_begin;
		if (n < actions.size()) {
			Plugin* plugin = 0;
			const Glib::ustring& plugin_name = actions[n]->plugin();
			if (! plugin_name.empty()) {
				std::map<const Glib::ustring, Plugin*>::iterator plugin_i
					= _plugins.find(plugin_name);
				if (plugin_i != _plugins.end()) {
					plugin = plugin_i->second;
				}

			}
			if (actions[n]->state() == normal) {
				_control.free_canvas();
				actions[n]->mark_state(active);

				if (plugin) {
					_control.action_activate(plugin, actions[n]->plugin_arg());
				}

				_configuration->audio()->play(Audio::action_select);

			} else {
				_control.action_deactivate();
				actions[n]->mark_state(normal);
			}
		}
	} else if (   _control.gamestate() >= STATE_question
	           && _control.gamestate() <= STATE_show_answer
	           && command >= command_next_status_begin
	           && command <= command_set_status_end) {
		if (command <= command_next_status_end) {
			const unsigned int n = command - command_next_status_begin;
			_control.next_status(n);
		} else if (   command >= command_prev_status_begin
		           && command <= command_prev_status_end) {
			const unsigned int n = command - command_prev_status_begin;
			_control.prev_status(n);
		} else if (   command >= command_set_status_begin
		           && command <= command_set_status_end) {
			const unsigned int N = command - command_set_status_begin;
			const unsigned int mod
				= (command_set_status_0_end - command_set_status_begin + 1);
			const unsigned int number = N / mod;
			const unsigned int n = N % mod;

			_control.set_status(number, n);
		}
	} else if (   _control.gamestate() >= STATE_question
	           && _control.gamestate() <= STATE_show_answer
	           && command >= command_select_action_autonomous_begin
	           && command <= command_select_action_autonomous_end) {
		const std::vector<Glib::ustring>::size_type n
			= command - command_select_action_autonomous_begin;
		if (n < _configuration->actions_autonomous().size()) {
			const Glib::ustring& plugin_name =
				_configuration->actions_autonomous()[n];
			if (! plugin_name.empty()) {
				std::map<const Glib::ustring, Plugin*>::iterator plugin_i
					= _plugins.find(plugin_name);
				if (plugin_i != _plugins.end()) {
					_control.free_canvas();
					_control.action_activate(plugin_i->second,
						_configuration->action_args_autonomous()[n]);
				}
			}
		}
	} else {
		switch (command) {
		case command_prev_question:
			if (_control.prev_question()) {
				_control.set_gamestate(STATE_question);
			}
			break;
		case command_next_question:
			if (_control.next_question()) {
				_control.set_gamestate(STATE_question);
			}
			break;
		case command_prev_level:
			_control.prev_level();
			break;
		case command_next_level:
			_control.next_level();
			break;
		case command_toggle_fullscreen:
			if (_is_fullscreen) {
				unfullscreen();
			} else {
				fullscreen();
			}
			break;
		case command_quit:
			if (_is_fullscreen) {
				unfullscreen();
			} else {
				Gtk::Main::quit();
			}
			break;
		case command_continue:
		case command_continue_benevolent:
			if (_timer_connection) {
				_timer_connection.disconnect();
			}
			if (_control.substate_next()) {
				break;
			}

			switch (_control.gamestate()) {
			case STATE_question:
				if (_control.answer_selected(_control.correct())) {
					_control.set_gamestate(STATE_answer);
				} else {
					std::cerr << _("Cannot mark correct answer.") << '\n';
				}

				break;
			case STATE_answer: {
				std::vector<ItemAnswer*> answers =
					_configuration->item_top()->answers();
				if (command == command_continue_benevolent) {
					_control.answer_all_correct();
					_control.set_gamestate(STATE_show_answer);
				} else {
					try {
						if (_control.answer_correct_correct()) {
							_control.set_gamestate(STATE_show_answer);
						} else {
							_control.set_gamestate(STATE_game_lost);
						}
					} catch (const ParserTriviaError& e) {
						std::cerr << e.what() << '\n';
					}
				}

				break;
			}
			case STATE_show_answer:
				if (   ! _control.next_level()
				    || ! _control.next_question()) {
					_control.set_gamestate(STATE_end);

					_configuration->item_top()->set_mode(mode_outro_win);
					_configuration->audio()->play(Audio::outro_win);
				} else {
					_control.set_gamestate(STATE_question);
				}
				break;
			case STATE_game_lost: {
				int seconds =
					_configuration->item_top()->set_mode(mode_outro_lose);
				_configuration->audio()->play(Audio::outro_lose);
				_control.set_gamestate(STATE_end);

				if (seconds > 0) {
					_timer_connection = Glib::signal_timeout().connect(
						sigc::mem_fun(*this, &Quimeleon::auto_continue),
						seconds * 1000);
				}
				break;
			}
			case STATE_end: {
				int seconds =
					_configuration->item_top()->set_sequence_pos(
						_configuration->item_top()->sequence_pos() + 1);
				if (seconds < 0) {
					_configuration->item_top()->set_mode(mode_black);

					_configuration->audio()->mute();
					_control.set_gamestate(STATE_black);
				} else if (seconds > 0) {
					_timer_connection = Glib::signal_timeout().connect(
						sigc::mem_fun(*this, &Quimeleon::auto_continue),
						seconds * 1000);
				}
				break;
			}
			case STATE_intro: {
				int seconds =
					_configuration->item_top()->set_sequence_pos(
						_configuration->item_top()->sequence_pos() + 1);
				if (seconds < 0) {
					const std::vector<ItemAction*>& actions =
						_configuration->item_top()->actions();
					for (std::vector<ItemAction*>::const_iterator i
					     	= actions.begin(); i < actions.end(); i++) {
						(*i)->mark_state(normal);
					}
					_control.reset_status();
					_control.reset_question();
					_control.reset_level();
					_configuration->item_top()->set_mode(mode_blank);

					_control.set_gamestate(STATE_start);
				} else if (seconds > 0) {
					_timer_connection = Glib::signal_timeout().connect(
						sigc::mem_fun(*this, &Quimeleon::auto_continue),
						seconds * 1000);
				}
				break;
			}
			case STATE_start:
				_control.set_question(0);
				_configuration->item_top()->set_mode(mode_trivia);
				_control.set_gamestate(STATE_question);
				break;
			case STATE_black: {
				int seconds = _configuration->item_top()->set_mode(mode_intro);
				_control.set_gamestate(STATE_intro);

				if (seconds > 0) {
					_timer_connection = Glib::signal_timeout().connect(
						sigc::mem_fun(*this, &Quimeleon::auto_continue),
						seconds * 1000);
				}
				break;
			}
			default: {}
			}

			break;
		default:
			return false;
		}
	}
	} catch (const Exception& e) {
		std::cerr << e.what() << '\n';
		_display.get_window()->set_cursor();
		Gtk::MessageDialog dialog(Glib::locale_to_utf8(e.what()), false,
		                          Gtk::MESSAGE_ERROR, Gtk::BUTTONS_NONE);
		dialog.add_button(Gtk::Stock::QUIT, Gtk::RESPONSE_OK);

		dialog.run();


		Gtk::Main::quit();
	}

	return true;
}




bool
Quimeleon::on_window_state_event(GdkEventWindowState* event) throw()
{
	if (event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) {
		if (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN) {
			_is_fullscreen = true;
			_menu->hide();
			_toolbar->hide();
			_display.get_window()->set_cursor(hidden_mouse_cursor());
		} else {
			_is_fullscreen = false;
			_menu->show();
			_toolbar->show();
			_display.get_window()->set_cursor();
		}
	}


	return true;
}



bool
Quimeleon::auto_continue()
{
	switch (_control.gamestate()) {
	case STATE_end: {
		int seconds = _configuration->item_top()->set_sequence_pos(
			_configuration->item_top()->sequence_pos() + 1);
		if (seconds < 0) {
			_configuration->item_top()->set_mode(mode_black);

			_configuration->audio()->mute();
			_control.set_gamestate(STATE_black);
		} else if (seconds > 0) {
			_timer_connection = Glib::signal_timeout().connect(
				sigc::mem_fun(*this, &Quimeleon::auto_continue),
					seconds * 1000);
		}
		break;
	}
	case STATE_intro: {
		int seconds = _configuration->item_top()->set_sequence_pos(
			_configuration->item_top()->sequence_pos() + 1);
		if (seconds < 0) {
			const std::vector<ItemAction*>& actions =
				_configuration->item_top()->actions();
			for (std::vector<ItemAction*>::const_iterator i
			     	= actions.begin(); i < actions.end(); i++) {
				(*i)->mark_state(normal);
			}
			_control.reset_status();
			_control.reset_question();
			_control.reset_level();
			_configuration->item_top()->set_mode(mode_blank);

			_control.set_gamestate(STATE_start);
		} else if (seconds > 0) {
			_timer_connection = Glib::signal_timeout().connect(
				sigc::mem_fun(*this, &Quimeleon::auto_continue),
				seconds * 1000);
		}
		break;
	}
	default:
		break;
	}

	return false;
}



bool
Quimeleon::start_mouse_watch(GdkEventCrossing* event)
{
	Gdk::ModifierType dummy_mask;
	_display.get_window()->get_pointer(_mouse_x, _mouse_y, dummy_mask);

	_connection_mouse = Glib::signal_timeout().connect(
		sigc::mem_fun(this, &Quimeleon::timer_mouse), 1000);

	return true;
}



bool
Quimeleon::timer_mouse()
{
	Gdk::ModifierType dummy_mask;
	int mouse_x;
	int mouse_y;
	_display.get_window()->get_pointer(mouse_x, mouse_y, dummy_mask);

	if (   mouse_x - _mouse_x < 2
	    && mouse_y - _mouse_y < 2) {
		_display.get_window()->set_cursor(hidden_mouse_cursor());

		_connection_mouse = _display.signal_motion_notify_event().connect(
			sigc::mem_fun(this, &Quimeleon::on_mouse_motion_notify_event));

		return false;
	} else {
		_mouse_x = mouse_x;
		_mouse_y = mouse_y;

		return true;
	}
}



bool
Quimeleon::on_mouse_motion_notify_event(GdkEventMotion* event)
{
	_connection_mouse.disconnect();

	if (! _is_fullscreen) {
		_display.get_window()->set_cursor();
		_connection_mouse = Glib::signal_timeout().connect(
			sigc::mem_fun(this, &Quimeleon::timer_mouse), 1000);
	}

	return false;
}



bool
Quimeleon::stop_mouse_watch(GdkEventCrossing* event)
{
	if (_connection_mouse) {
		_connection_mouse.disconnect();
	}
	_display.get_window()->set_cursor();

	return true;
}



void
Quimeleon::on_show()
{
	Window::on_show();
	if (! _is_fullscreen
	    && _configuration->do_show_hint()) {
		Welcome welcome;
		welcome.run();

		if (! welcome.do_show_hint()) {
			_configuration->set_do_show_hint(false);
		}
		try {
			_configuration->save_options();
		} catch (const Exception& e) {
			Gtk::MessageDialog(welcome,
				Glib::locale_to_utf8(e.what()),
				false, Gtk::MESSAGE_ERROR).run();
		}
	}
}





} // end namespace quimeleon
